@@ -27,6 +27,7 @@
 
 #include <vpn.h>
 #include <worker.h>
+#include <main.h>
 
 const char *human_addr(const struct sockaddr *sa, socklen_t salen,
 		       void *_buf, size_t buflen)
@@ -42,8 +43,7 @@ const char *human_addr(const struct sockaddr *sa, socklen_t salen,
 	buf++;
 	buflen--;
 
-	if (getnameinfo(sa, salen, buf, buflen, NULL, 0, NI_NUMERICHOST) !=
-	    0)
+	if (getnameinfo(sa, salen, buf, buflen, NULL, 0, NI_NUMERICHOST) != 0)
 		return NULL;
 
 	l = strlen(buf);
@@ -58,8 +58,7 @@ const char *human_addr(const struct sockaddr *sa, socklen_t salen,
 	buf++;
 	buflen--;
 
-	if (getnameinfo(sa, salen, NULL, 0, buf, buflen, NI_NUMERICSERV) !=
-	    0)
+	if (getnameinfo(sa, salen, NULL, 0, buf, buflen, NI_NUMERICSERV) != 0)
 		return NULL;
 
 	return save_buf;
@@ -122,3 +121,24 @@ void __attribute__ ((format(printf, 4, 5)))
 	
 	return;
 }
+
+void  mslog_hex(const main_server_st * s, const struct proc_st* proc,
+    	int priority, const char *prefix, uint8_t* bin, unsigned bin_size)
+{
+	char buf[512];
+	int ret;
+	size_t buf_size;
+	gnutls_datum_t data = {bin, bin_size};
+	
+	if (priority == LOG_DEBUG && s->config->debug == 0)
+		return;
+
+	buf_size = sizeof(buf);
+	ret = gnutls_hex_encode(&data, buf, &buf_size);
+	if (ret < 0)
+		return;
+	
+	mslog(s, proc, priority, "%s %s", prefix, buf);
+
+	return;
+}
