@@ -28,13 +28,17 @@
 #include <fcntl.h>
 #include <sys/socket.h>
 #include <netdb.h>
-#include <signal.h>
+#include <system.h>
 #include <errno.h>
 #include <sys/ioctl.h>
 #include <gnutls/gnutls.h>
 #include <gnutls/crypto.h>
 #include <tlslib.h>
+#include <sys/un.h>
+#include <cloexec.h>
 #include "ipc.h"
+#include "setproctitle.h"
+#include <sec-mod.h>
 
 #include <vpn.h>
 #include <cookies.h>
@@ -108,21 +112,105 @@ int send_udp_fd(main_server_st* s, struct proc_st * proc, int fd)
 	return(sendmsg(proc->fd, &hdr, 0));
 }
 
+int handle_script_exit(main_server_st *s, struct proc_st* proc, int code)
+{
+int ret;
+
+	if (code == 0) {
+		ret = send_auth_reply(s, proc, REP_AUTH_OK);
+		if (ret < 0) {
+			mslog(s, proc, LOG_ERR, "could not send reply auth cmd.");
+			ret = ERR_BAD_COMMAND;
+			goto fail;
+		}
+	} else {
+		mslog(s, proc, LOG_INFO, "failed authentication attempt for user '%s'", proc->username);
+		ret = send_auth_reply( s, proc, REP_AUTH_FAILED);
+		if (ret < 0) {
+			mslog(s, proc, LOG_ERR, "could not send reply auth cmd.");
+			ret = ERR_BAD_COMMAND;
+			goto fail;
+		}
+	}
+	ret = 0;
+
+fail:	
+	/* we close the lease tun fd both on success and failure.
+	 * The parent doesn't need to keep the tunfd.
+	 */
+	if (proc->lease) {
+		if (proc->lease->fd >= 0)
+			close(proc->lease->fd);
+		proc->lease->fd = -1;
+	}
+
+	return ret;
+}
+
+static int accept_user(main_server_st *s, struct proc_st* proc, unsigned cmd)
+{
+int ret;
+const char* group;
+
+	mslog(s, proc, LOG_DEBUG, "accepting user '%s'", proc->username);
+	proc_auth_deinit(s, proc);
+
+	/* check for multiple connections */
+	ret = check_multiple_users(s, proc);
+	if (ret < 0) {
+		mslog(s, proc, LOG_INFO, "user '%s' tried to connect more than %u times", proc->username, s->config->max_same_clients);
+		return ret;
+	}
+
+	ret = open_tun(s, proc);
+	if (ret < 0) {
+		return -1;
+	}
+
+	if (proc->groupname[0] == 0)
+		group = "[unknown]";
+	else
+		group = proc->groupname;
+
+	if (cmd == AUTH_REQ || cmd == AUTH_INIT) {
+		/* generate cookie */
+		ret = generate_cookie(s, proc);
+		if (ret < 0) {
+			return ERR_BAD_COMMAND;
+		}
+		mslog(s, proc, LOG_INFO, "user '%s' of group '%s' authenticated", proc->username, group);
+	} else {
+		mslog(s, proc, LOG_INFO, "user '%s' of group '%s' re-authenticated (using cookie)", proc->username, group);
+	}
+
+	/* do scripts and utmp */
+	ret = user_connected(s, proc);
+	if (ret == ERR_WAIT_FOR_SCRIPT) {
+		return 0;
+	}
+
+	if (ret < 0) {
+		mslog(s, proc, LOG_INFO, "user '%s' disconnected due to script", proc->username);
+	}
+	
+	return 0;
+}
+
 int handle_commands(main_server_st *s, struct proc_st* proc)
 {
 	struct iovec iov[2];
 	uint8_t cmd;
 	struct msghdr hdr;
-	struct lease_st *lease;
+	/* FIXME: do not write directly to the union */
 	union {
 		struct cmd_auth_req_st auth;
 		struct cmd_auth_cookie_req_st cauth;
 		struct cmd_resume_store_req_st sresume;
 		struct cmd_resume_fetch_req_st fresume;
 		struct cmd_tun_mtu_st tmtu;
+		struct cmd_auth_init_st auth_init;
 	} cmd_data;
 	int ret, cmd_data_len, e;
-	const char* group;
 
 	memset(&cmd_data, 0, sizeof(cmd_data));
 	
@@ -148,21 +236,20 @@ int handle_commands(main_server_st *s, struct proc_st* proc)
 	}
 
 	cmd_data_len = ret - 1;
-	
+
 	switch(cmd) {
 		case CMD_TUN_MTU:
 			if (cmd_data_len != sizeof(cmd_data.tmtu)) {
 				mslog(s, proc, LOG_ERR, "error in received message (cmd %u) length.", (unsigned)cmd);
-				return -2;
+				return ERR_BAD_COMMAND;
 			}
 			
 			set_tun_mtu(s, proc, cmd_data.tmtu.mtu);
 			break;
-
 		case RESUME_STORE_REQ:
 			if (cmd_data_len <= sizeof(cmd_data.sresume)-MAX_SESSION_DATA_SIZE) {
 				mslog(s, proc, LOG_ERR, "error in received message (cmd %u) length.", (unsigned)cmd);
-				return -2;
+				return ERR_BAD_COMMAND;
 			}
 			ret = handle_resume_store_req(s, proc, &cmd_data.sresume);
 			if (ret < 0) {
@@ -174,7 +261,7 @@ int handle_commands(main_server_st *s, struct proc_st* proc)
 		case RESUME_DELETE_REQ:
 			if (cmd_data_len != sizeof(cmd_data.fresume)) {
 				mslog(s, proc, LOG_ERR, "error in received message (cmd %u) length.", (unsigned)cmd);
-				return -2;
+				return ERR_BAD_COMMAND;
 			}
 			ret = handle_resume_delete_req(s, proc, &cmd_data.fresume);
 			if (ret < 0) {
@@ -187,7 +274,7 @@ int handle_commands(main_server_st *s, struct proc_st* proc)
 
 			if (cmd_data_len != sizeof(cmd_data.fresume)) {
 				mslog(s, proc, LOG_ERR, "error in received message (%u) length.", (unsigned)cmd);
-				return -2;
+				return ERR_BAD_COMMAND;
 			}
 			ret = handle_resume_fetch_req(s, proc, &cmd_data.fresume, &reply);
 			if (ret < 0) {
@@ -199,103 +286,109 @@ int handle_commands(main_server_st *s, struct proc_st* proc)
 			
 			if (ret < 0) {
 				mslog(s, proc, LOG_ERR, "could not send reply cmd %d.", (unsigned) cmd);
-				return -2;
+				return ERR_BAD_COMMAND;
+			}
+			
+			break;
+
+		case AUTH_INIT:
+			if (cmd_data_len != sizeof(cmd_data.auth_init)) {
+				mslog(s, proc, LOG_ERR, "error in received message (%u) length.", (unsigned)cmd);
+				return ERR_BAD_COMMAND;
+			}
+
+			if (proc->auth_status != PS_AUTH_INACTIVE) {
+				mslog(s, proc, LOG_ERR, "received authentication init when complete.");
+				return ERR_BAD_COMMAND;
+			}
+
+			ret = handle_auth_init(s, proc, &cmd_data.auth_init);
+			if (ret == ERR_AUTH_CONTINUE) {
+				proc->auth_status = PS_AUTH_INIT;
+
+				ret = send_auth_reply_msg(s, proc);
+				if (ret < 0) {
+					mslog(s, proc, LOG_ERR, "could not send reply auth cmd.");
+					return ret;
+				}
+				break; /* wait for another command */
+			} else if (ret < 0) {
+				add_to_ip_ban_list(s, &proc->remote_addr, proc->remote_addr_len);
+				goto cleanup;
 			}
 			
 			break;
 
 		case AUTH_REQ:
-		case AUTH_COOKIE_REQ:
-			lease = NULL;
 
-			if (cmd == AUTH_REQ) {
-				if (cmd_data_len != sizeof(cmd_data.auth)) {
-					mslog(s, proc, LOG_ERR, "error in received message (%u) length.", (unsigned)cmd);
-					return -2;
-				}
-
-				ret = handle_auth_req(s, proc, &cmd_data.auth, &lease);
-			} else {
-				if (cmd_data_len != sizeof(cmd_data.cauth)) {
-					mslog(s, proc, LOG_ERR, "error in received message (%u) length.", (unsigned)cmd);
-					return -2;
-				}
-
-				ret = handle_auth_cookie_req(s, proc, &cmd_data.cauth, &lease);
+			if (proc->auth_status != PS_AUTH_INIT) {
+				mslog(s, proc, LOG_ERR, "received authentication request when not initialized.");
+				return ERR_BAD_COMMAND;
 			}
 
-			if (ret == 0) {
-				/* check for multiple connections */
-				ret = check_multiple_users(s, proc);
-				if (ret < 0) {
-					mslog(s, proc, LOG_INFO, "user '%s' tried to connect more than %u times", proc->username, s->config->max_same_clients);
-				}
+			if (cmd_data_len != sizeof(cmd_data.auth)) {
+				mslog(s, proc, LOG_ERR, "error in received message (%u) length.", (unsigned)cmd);
+				return ERR_BAD_COMMAND;
+			}
+			
+			proc->auth_reqs++;
+			if (proc->auth_reqs > MAX_AUTH_REQS) {
+				mslog(s, proc, LOG_ERR, "received too many authentication requests.");
+				return ERR_BAD_COMMAND;
+			}
 
-				/* do scripts and utmp */
-				if (ret == 0) {
-					ret = user_connected(s, proc, lease);
-					if (ret < 0) {
-						mslog(s, proc, LOG_INFO, "user '%s' disconnected due to script", proc->username);
-					}
+			ret = handle_auth_req(s, proc, &cmd_data.auth);
+			if (ret == ERR_AUTH_CONTINUE) {
+				ret = send_auth_reply_msg(s, proc);
+				if (ret < 0) {
+					mslog(s, proc, LOG_ERR, "could not send reply auth cmd.");
+					return ret;
 				}
-			} else {
+				break; /* wait for another command */
+			} else if (ret < 0) {
 				add_to_ip_ban_list(s, &proc->remote_addr, proc->remote_addr_len);
+				goto cleanup;
 			}
 
-			if (ret == 0) {
-				if (proc->groupname[0] == 0)
-					group = "[unknown]";
-				else
-					group = proc->groupname;
+			ret = accept_user(s, proc, cmd);
+			if (ret < 0) {
+				goto cleanup;
+			}
+			proc->auth_status = PS_AUTH_COMPLETED;
+			goto cleanup;
 
-				if (cmd == AUTH_REQ) {
-					/* generate and store cookie */
-					ret = generate_and_store_vals(s, proc);
-					if (ret < 0) {
-						ret = -2;
-						goto lease_cleanup;
-					}
-					mslog(s, proc, LOG_INFO, "user '%s' of group '%s' authenticated", proc->username, group);
-				} else {
-					mslog(s, proc, LOG_INFO, "user '%s' of group '%s' re-authenticated (using cookie)", proc->username, group);
-				}
-				
-				ret = send_auth_reply(s, proc, REP_AUTH_OK, lease);
-				if (ret < 0) {
-					mslog(s, proc, LOG_ERR, "could not send reply cmd %d.", (unsigned)cmd);
-					ret = -2;
-					goto lease_cleanup;
-				}
-
-				proc->lease = lease;
-				proc->lease->in_use = 1;
-				ret = 0;
-			} else {
-				mslog(s, proc, LOG_INFO, "failed authentication attempt for user '%s'", proc->username);
-				ret = send_auth_reply( s, proc, REP_AUTH_FAILED, NULL);
-				if (ret < 0) {
-					mslog(s, proc, LOG_ERR, "could not send reply cmd.");
-					ret = -2;
-					goto lease_cleanup;
-				}
-				ret = 0;
+		case AUTH_COOKIE_REQ:
+			
+			if (cmd_data_len != sizeof(cmd_data.cauth)) {
+				mslog(s, proc, LOG_ERR, "error in received message (%u) length.", (unsigned)cmd);
+				return ERR_BAD_COMMAND;
 			}
 
-lease_cleanup:
-			/* we close the lease tun fd both on success and failure.
-			 * The parent doesn't need to know the tunfd.
-			 */
-			if (lease) {
-				if (lease->fd >= 0)
-					close(lease->fd);
-				lease->fd = -1;
+			if (proc->auth_status != PS_AUTH_INACTIVE) {
+				mslog(s, proc, LOG_ERR, "received unexpected cookie authentication.");
+				return ERR_BAD_COMMAND;
 			}
-			return ret;
 
-			break;
+			ret = handle_auth_cookie_req(s, proc, &cmd_data.cauth);
+			if (ret < 0) {
+				add_to_ip_ban_list(s, &proc->remote_addr, proc->remote_addr_len);
+				goto cleanup;
+			}
+
+			ret = accept_user(s, proc, cmd);
+			if (ret < 0) {
+				goto cleanup;
+			}
+
+			proc->auth_status = PS_AUTH_COMPLETED;
+
+cleanup:
+			/* no script was called. Handle it as a successful script call. */
+			return handle_script_exit(s, proc, ret);
+
 		default:
 			mslog(s, proc, LOG_ERR, "unknown CMD 0x%x.", (unsigned)cmd);
-			return -2;
+			return ERR_BAD_COMMAND;
 	}
 	
 	return 0;
@@ -364,3 +457,36 @@ struct banned_st *btmp;
 	
 	list_add(&s->ban_list.head, &(btmp->list));
 }
+
+void run_sec_mod(main_server_st * s)
+{
+int e;
+pid_t pid;
+char file[_POSIX_PATH_MAX];
+const char *p;
+
+	/* make socket name */
+	snprintf(s->socket_file, sizeof(s->socket_file), "%s.%u", s->config->socket_file_prefix, (unsigned)getpid());
+	p = s->socket_file;
+	if (s->config->chroot_dir != NULL) {
+		snprintf(file, sizeof(file), "%s/%s.%u", 
+			s->config->chroot_dir, s->config->socket_file_prefix, (unsigned)getpid());
+		p = file;
+	}
+
+	pid = fork();
+	if (pid == 0) { /* child */
+		clear_lists(s);
+		kill_on_parent_kill(SIGTERM);
+		setproctitle(PACKAGE_NAME"-secmod");
+
+		sec_mod_server(s->config, p);
+		exit(0);
+	} else if (pid > 0) { /* parent */
+		s->sec_mod_pid = pid;
+	} else {
+		e = errno;
+		mslog(s, NULL, LOG_ERR, "error in fork(): %s", strerror(e));
+		exit(1);
+	}
+}
