@@ -28,12 +28,15 @@
 #include <fcntl.h>
 #include <sys/socket.h>
 #include <netdb.h>
-#include <signal.h>
+#include <system.h>
 #include <errno.h>
 #include <sys/ioctl.h>
+#include <sys/resource.h>
 #include <cloexec.h>
+#include <script-list.h>
 
 #include <gnutls/x509.h>
+#include <gnutls/crypto.h>
 #include <tlslib.h>
 #include "ipc.h"
 #include "setproctitle.h"
@@ -42,6 +45,7 @@
 #endif
 
 #include <main.h>
+#include <main-auth.h>
 #include <worker.h>
 #include <cookies.h>
 #include <tun.h>
@@ -50,7 +54,7 @@
 int syslog_open = 0;
 static unsigned int terminate = 0;
 static unsigned int reload_conf = 0;
-unsigned int need_maintainance = 0;
+unsigned int need_maintenance = 0;
 static unsigned int need_children_cleanup = 0;
 
 static void ms_sleep(unsigned ms)
@@ -119,7 +123,7 @@ int _listen_ports(struct cfg_st* config, struct addrinfo *res, struct listen_lis
 		if (ptr->ai_socktype == SOCK_DGRAM) {
 #if defined(IP_DONTFRAG)
 			y = 1;
-			if (setsockopt(s, IPPROTO_IP, IP_DONTFRAG,
+			if (setsockopt(s, SOL_IP, IP_DONTFRAG,
 				       (const void *) &y, sizeof(y)) < 0)
 				perror("setsockopt(IP_DF) failed");
 #elif defined(IP_MTU_DISCOVER)
@@ -285,17 +289,63 @@ int s, y, e;
 	return 0;
 }
 
+static void remove_proc(main_server_st* s, struct proc_st *proc, unsigned k)
+{
+	if (k)
+		kill(proc->pid, SIGTERM);
+
+	user_disconnected(s, proc);
+
+	/* close the intercomm fd */
+	if (proc->fd >= 0)
+		close(proc->fd);
+	proc->fd = -1;
+	proc->pid = -1;
+	
+	if (proc->auth_ctx != NULL)
+		proc_auth_deinit(s, proc);
+
+	if (proc->lease)
+		proc->lease->in_use = 0;
+	list_del(&proc->list);
+	free(proc);
+	s->active_clients--;
+}
 
 static void cleanup_children(main_server_st *s)
 {
-int status;
+int status, estatus, ret;
 pid_t pid;
+struct script_wait_st *stmp, *spos;
 
 	while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
-		if (WEXITSTATUS(status) != 0 ||
-			(WIFSIGNALED(status) && WTERMSIG(status) == SIGSEGV)) {
-			if (WIFSIGNALED(status))
+		estatus = WEXITSTATUS(status);
+		
+		if (pid == s->sec_mod_pid) {
+			mslog(s, NULL, LOG_ERR, "ocserv-secmod died unexpectedly");
+			terminate = 1;
+		}
+
+		/* check if someone was waiting for that pid */
+		list_for_each_safe(&s->script_list.head, stmp, spos, list) {
+			if (stmp->pid == pid) {
+				mslog(s, stmp->proc, LOG_DEBUG, "%s-script exit status: %u", stmp->up?"connect":"disconnect", estatus);
+				ret = handle_script_exit(s, stmp->proc, estatus);
+				if (ret < 0)
+					remove_proc(s, stmp->proc, 0);
+				list_del(&stmp->list);
+				free(stmp);
+				break;
+			}
+		}
+	
+		if (WIFSIGNALED(status)) {
+			if (WTERMSIG(status) == SIGSEGV)
 				mslog(s, NULL, LOG_ERR, "Child %u died with sigsegv\n", (unsigned)pid);
+			else if (WTERMSIG(status) == SIGSYS)
+				mslog(s, NULL, LOG_ERR, "Child %u died with sigsys\n", (unsigned)pid);
+			else
+				mslog(s, NULL, LOG_ERR, "Child %u died with signal %d\n", (unsigned)pid, (int)WTERMSIG(status));
 		}
 	}
 	need_children_cleanup = 0;
@@ -308,21 +358,20 @@ static void handle_children(int signo)
 
 static void handle_alarm(int signo)
 {
-	need_maintainance = 1;
+	need_maintenance = 1;
 }
 
-
-
 static void drop_privileges(main_server_st* s)
 {
 	int ret, e;
+	struct rlimit rl;
 
 	if (s->config->chroot_dir) {
 		chdir(s->config->chroot_dir);
 		ret = chroot(s->config->chroot_dir);
 		if (ret != 0) {
 			e = errno;
-			mslog(s, NULL, LOG_ERR, "Cannot chroot to %s: %s", s->config->chroot_dir, strerror(e));
+			mslog(s, NULL, LOG_ERR, "cannot chroot to %s: %s", s->config->chroot_dir, strerror(e));
 			exit(1);
 		}
 	}
@@ -331,7 +380,7 @@ static void drop_privileges(main_server_st* s)
 		ret = setgid(s->config->gid);
 		if (ret < 0) {
 			e = errno;
-			mslog(s, NULL, LOG_ERR, "Cannot set gid to %d: %s\n",
+			mslog(s, NULL, LOG_ERR, "cannot set gid to %d: %s\n",
 			       (int) s->config->gid, strerror(e));
 			exit(1);
 
@@ -342,12 +391,35 @@ static void drop_privileges(main_server_st* s)
 		ret = setuid(s->config->uid);
 		if (ret < 0) {
 			e = errno;
-			mslog(s, NULL, LOG_ERR, "Cannot set uid to %d: %s\n",
+			mslog(s, NULL, LOG_ERR, "cannot set uid to %d: %s\n",
 			       (int) s->config->uid, strerror(e));
 			exit(1);
 
 		}
 	}
+
+	rl.rlim_cur = 0;
+	rl.rlim_max = 0;
+	ret = setrlimit(RLIMIT_NPROC, &rl);
+	if (ret < 0) {
+		e = errno;
+		mslog(s, NULL, LOG_ERR, "cannot enforce NPROC limit: %s\n",
+		       strerror(e));
+	}
+
+#if 0
+#define MAX_WORKER_MEM (4*1024*1024)
+	if (s->config->debug == 0) {
+		rl.rlim_cur = MAX_WORKER_MEM;
+		rl.rlim_max = MAX_WORKER_MEM;
+		ret = setrlimit(RLIMIT_AS, &rl);
+		if (ret < 0) {
+			e = errno;
+			mslog(s, NULL, LOG_ERR, "cannot enforce AS limit: %s\n",
+			       strerror(e));
+		}
+	}
+#endif
 }
 
 /* clears the server llist and clist. To be used after fork().
@@ -357,6 +429,7 @@ void clear_lists(main_server_st *s)
 {
 	struct listener_st *ltmp, *lpos;
 	struct proc_st *ctmp, *cpos;
+	struct script_wait_st *script_tmp, *script_pos;
 	struct banned_st *btmp, *bpos;
 
 	list_for_each_safe(&s->llist.head, ltmp, lpos, list) {
@@ -370,6 +443,7 @@ void clear_lists(main_server_st *s)
 		if (ctmp->fd >= 0)
 			close(ctmp->fd);
 		list_del(&ctmp->list);
+		memset(ctmp, 0, sizeof(*ctmp));
 		free(ctmp);
 		s->clist.total--;
 	}
@@ -379,6 +453,11 @@ void clear_lists(main_server_st *s)
 		free(btmp);
 	}
 
+	list_for_each_safe(&s->script_list.head, script_tmp, script_pos, list) {
+		list_del(&script_tmp->list);
+		free(script_tmp);
+	}
+
 	tls_cache_deinit(s->tls_db);
 }
 
@@ -386,6 +465,8 @@ static void kill_children(main_server_st* s)
 {
 	struct proc_st *ctmp;
 
+	/* kill the security module server */
+	kill(s->sec_mod_pid, SIGTERM);
 	list_for_each(&s->clist.head, ctmp, list) {
 		if (ctmp->pid != -1) {
 			kill(ctmp->pid, SIGTERM);
@@ -394,20 +475,6 @@ static void kill_children(main_server_st* s)
 	}
 }
 
-static void remove_proc(struct proc_st *ctmp)
-{
-	/* close the intercomm fd */
-	if (ctmp->fd >= 0)
-		close(ctmp->fd);
-	ctmp->fd = -1;
-	ctmp->pid = -1;
-
-	if (ctmp->lease)
-		ctmp->lease->in_use = 0;
-	list_del(&ctmp->list);
-	free(ctmp);
-}
-
 static void handle_term(int signo)
 {
 	/* kill all children */
@@ -438,7 +505,7 @@ int connected = 0;
 	cli_addr_size = sizeof(cli_addr);
 	ret = recvfrom(listener->fd, buffer, sizeof(buffer), MSG_PEEK, (void*)&cli_addr, &cli_addr_size);
 	if (ret < 0) {
-		mslog(s, NULL, LOG_INFO, "Error receiving in UDP socket");
+		mslog(s, NULL, LOG_INFO, "error receiving in UDP socket");
 		return -1;
 	}
 	
@@ -453,11 +520,11 @@ int connected = 0;
 	mslog(s, NULL, LOG_DEBUG, "DTLS hello version: %u.%u", (unsigned int)buffer[RECORD_PAYLOAD_POS], (unsigned int)buffer[RECORD_PAYLOAD_POS+1]);
 	if (buffer[1] != 254 && (buffer[1] != 1 && buffer[2] != 0) &&
 		buffer[RECORD_PAYLOAD_POS] != 254 && (buffer[RECORD_PAYLOAD_POS] != 0 && buffer[RECORD_PAYLOAD_POS+1] != 0)) {
-		mslog(s, NULL, LOG_INFO, "Unknown DTLS version: %u.%u", (unsigned)buffer[1], (unsigned)buffer[2]);
+		mslog(s, NULL, LOG_INFO, "unknown DTLS version: %u.%u", (unsigned)buffer[1], (unsigned)buffer[2]);
 		goto fail;
 	}
 	if (buffer[0] != 22) {
-		mslog(s, NULL, LOG_INFO, "Unexpected DTLS content type: %u", (unsigned int)buffer[0]);
+		mslog(s, NULL, LOG_INFO, "unexpected DTLS content type: %u", (unsigned int)buffer[0]);
 		goto fail;
 	}
 
@@ -466,8 +533,8 @@ int connected = 0;
 	session_id = &buffer[RECORD_PAYLOAD_POS+HANDSHAKE_SESSION_ID_POS+1];
 
 	/* search for the IP and the session ID in all procs */
-	list_for_each(&s->clist.head, ctmp, list) {
 
+	list_for_each(&s->clist.head, ctmp, list) {
 		if (session_id_size == ctmp->session_id_size &&
 			memcmp(session_id, ctmp->session_id, session_id_size) == 0) {
 
@@ -480,15 +547,14 @@ int connected = 0;
 
 			ret = send_udp_fd(s, ctmp, listener->fd);
 			if (ret < 0) {
-				mslog(s, ctmp, LOG_ERR, "Error passing UDP socket");
+				mslog(s, ctmp, LOG_ERR, "error passing UDP socket");
 				return -1;
 			}
-			mslog(s, ctmp, LOG_DEBUG, "Passed UDP socket");
+			mslog(s, ctmp, LOG_DEBUG, "passed UDP socket");
 			ctmp->udp_fd_received = 1;
 			connected = 1;
 			
 			reopen_udp_port(listener);
-
 			break;
 		}
 	}
@@ -512,31 +578,31 @@ static void check_other_work(main_server_st *s)
 	if (reload_conf != 0) {
 		mslog(s, NULL, LOG_INFO, "HUP signal was received; reloading configuration");
 		reload_cfg_file(s->config);
-		tls_global_init_certs(s);
 		reload_conf = 0;
 	}
 
+	if (need_children_cleanup != 0) {
+		cleanup_children(s);
+	}
+
 	if (terminate != 0) {
 		mslog(s, NULL, LOG_DEBUG, "termination signal received; waiting for children to die");
 		kill_children(s);
 		closelog();
+		remove(s->socket_file);
+		remove_pid_file();
 		while (waitpid(-1, NULL, 0) > 0);
 		exit(0);
 	}
 
 	/* Check if we need to expire any cookies */
-	if (need_maintainance != 0) {
-		need_maintainance = 0;
-		mslog(s, NULL, LOG_INFO, "Performing maintainance");
+	if (need_maintenance != 0) {
+		need_maintenance = 0;
+		mslog(s, NULL, LOG_INFO, "Performing maintenance");
 		expire_tls_sessions(s);
-		expire_cookies(s);
 		expire_banned(s);
 		alarm(MAINTAINANCE_TIME(s));
 	}
-		
-	if (need_children_cleanup != 0) {
-		cleanup_children(s);
-	}
 }
 
 #ifdef HAVE_LIBWRAP
@@ -569,25 +635,32 @@ int main(int argc, char** argv)
 	int cmd_fd[2];
 	struct worker_st ws;
 	struct cfg_st config;
-	unsigned active_clients = 0, set;
+	unsigned set;
 	main_server_st s;
 
 	memset(&s, 0, sizeof(s));
 
 	list_head_init(&s.clist.head);
 	list_head_init(&s.ban_list.head);
+	list_head_init(&s.script_list.head);
 	tun_st_init(&tun);
 	tls_cache_init(&s.tls_db);
 
-	signal(SIGINT, handle_term);
-	signal(SIGTERM, handle_term);
-	signal(SIGPIPE, SIG_IGN);
-	signal(SIGHUP, handle_reload);
-	signal(SIGCHLD, handle_children);
-	signal(SIGALRM, handle_alarm);
+	ocsignal(SIGINT, handle_term);
+	ocsignal(SIGTERM, handle_term);
+	ocsignal(SIGPIPE, SIG_IGN);
+	ocsignal(SIGHUP, handle_reload);
+	ocsignal(SIGCHLD, handle_children);
+	ocsignal(SIGALRM, handle_alarm);
 
 	/* Initialize GnuTLS */
 	tls_global_init(&s);
+	
+	ret = gnutls_rnd(GNUTLS_RND_RANDOM, s.cookie_key, sizeof(s.cookie_key));
+	if (ret < 0) {
+		fprintf(stderr, "Error in cookie key generation\n");
+		exit(1);
+	}
 
 	/* load configuration */
 	ret = cmd_parser(argc, argv, &config);
@@ -606,12 +679,8 @@ int main(int argc, char** argv)
 	s.config = &config;
 	s.tun = &tun;
 
-	ret = cookie_db_init(&s);
-	if (ret < 0) {
-		fprintf(stderr, "Could not initialize cookie database.\n");
-		exit(1);
-	}
-	
+	main_auth_init(&s);
+
 	/* Listen to network ports */
 	ret = listen_ports(&config, &s.llist, config.name);
 	if (ret < 0) {
@@ -625,15 +694,13 @@ int main(int argc, char** argv)
 		flags |= LOG_PERROR;
 #endif
 	openlog("ocserv", flags, LOG_DAEMON);
+	syslog_open = 1;
 #ifdef HAVE_LIBWRAP
 	allow_severity = LOG_DAEMON|LOG_INFO;
 	deny_severity = LOG_DAEMON|LOG_WARNING;
 #endif	
 	memset(&ws, 0, sizeof(ws));
 
-	/* Initialize GnuTLS */
-	tls_global_init_certs(&s);
-	
 	if (config.foreground == 0) {
 		if (daemon(0, 0) == -1) {
 			e = errno;
@@ -643,11 +710,14 @@ int main(int argc, char** argv)
 	}
 
 	write_pid_file();
+	
+	run_sec_mod(&s);
+
+	/* Initialize certificates */
+	tls_global_init_certs(&s);
 
 	alarm(MAINTAINANCE_TIME(&s));
 
-	syslog_open = 1;
-
 	for (;;) {
 		check_other_work(&s);
 
@@ -701,7 +771,7 @@ int main(int argc, char** argv)
 					continue;
 				}
 				set_cloexec_flag (fd, 1);
-				
+
 				/* Check if the client is on the banned list */
 				ret = check_if_banned(&s, &ws.remote_addr, ws.remote_addr_len);
 				if (ret < 0) {
@@ -711,9 +781,9 @@ int main(int argc, char** argv)
 					break;
 				}
 
-				if (config.max_clients > 0 && active_clients >= config.max_clients) {
+				if (config.max_clients > 0 && s.active_clients >= config.max_clients) {
 					close(fd);
-					mslog(&s, NULL, LOG_INFO, "Reached maximum client limit (active: %u)", active_clients);
+					mslog(&s, NULL, LOG_INFO, "Reached maximum client limit (active: %u)", s.active_clients);
 					break;
 				}
 
@@ -738,9 +808,9 @@ int main(int argc, char** argv)
 					 */
 					close(cmd_fd[0]);
 					clear_lists(&s);
-					erase_cookies(&s);
 
 					setproctitle(PACKAGE_NAME"-worker");
+					kill_on_parent_kill(SIGTERM);
 
 					ws.config = &config;
 					ws.cmd_fd = cmd_fd[1];
@@ -749,10 +819,6 @@ int main(int argc, char** argv)
 					ws.conn_fd = fd;
 					ws.creds = &s.creds;
 
-					ret = tls_global_init_client(&ws);
-					if (ret < 0)
-						exit(1);
-
 					/* Drop privileges after this point */
 					drop_privileges(&s);
 
@@ -762,6 +828,7 @@ int main(int argc, char** argv)
 fork_failed:
 					close(cmd_fd[0]);
 				} else { /* parent */
+					/* add_proc */
 					ctmp = calloc(1, sizeof(struct proc_st));
 					if (ctmp == NULL) {
 						kill(pid, SIGTERM);
@@ -775,7 +842,7 @@ fork_failed:
 					set_cloexec_flag (cmd_fd[0], 1);
 
 					list_add(&s.clist.head, &(ctmp->list));
-					active_clients++;
+					s.active_clients++;
 				}
 				close(cmd_fd[1]);
 				close(fd);
@@ -786,7 +853,7 @@ fork_failed:
 				/* connection on UDP port */
 				ret = forward_udp_to_owner(&s, ltmp);
 				if (ret < 0) {
-					mslog(&s, NULL, LOG_INFO, "Could not determine the owner of received UDP packet");
+					mslog(&s, NULL, LOG_INFO, "could not determine the owner of received UDP packet");
 				}
 
 				if (config.rate_limit_ms > 0)
@@ -799,13 +866,8 @@ fork_failed:
 			if (FD_ISSET(ctmp->fd, &rd)) {
 				ret = handle_commands(&s, ctmp);
 				if (ret < 0) {
-					if (ret == -2) {
-						/* received a bad command from worker */
-						kill(ctmp->pid, SIGTERM);
-					}
-					user_disconnected(&s, ctmp);
-					remove_proc(ctmp);
-					active_clients--;
+					remove_from_script_list(&s, ctmp);
+					remove_proc(&s, ctmp, (ret==ERR_BAD_COMMAND)?1:0);
 				}
 			}
 		}
