@@ -8,13 +8,15 @@
 #include <vpn.h>
 #include <tlslib.h>
 #include "ipc.h"
+#include <cookies.h>
 
 int cmd_parser (int argc, char **argv, struct cfg_st* config);
 void reload_cfg_file(struct cfg_st* config);
 void write_pid_file(void);
+void remove_pid_file(void);
 
 /* set to 1 to start cleaning up cookies, sessions etc. */
-extern unsigned int need_maintainance;
+extern unsigned int need_maintenance;
 
 struct listener_st {
 	struct list_node list;
@@ -32,12 +34,26 @@ struct listen_list_st {
 	unsigned int total;
 };
 
+struct script_wait_st {
+	struct list_node list;
+
+	pid_t pid;
+	unsigned int up; /* connect or disconnect script */
+	struct proc_st* proc;
+};
+
+enum {
+	PS_AUTH_INACTIVE,
+	PS_AUTH_INIT,
+	PS_AUTH_COMPLETED,
+};
+
 struct proc_st {
 	struct list_node list;
 	int fd;
 	pid_t pid;
 	unsigned udp_fd_received; /* if the corresponding process has received a UDP fd */
-
+	
 	/* the tun lease this process has */
 	struct lease_st* lease;
 
@@ -55,6 +71,10 @@ struct proc_st {
 	char groupname[MAX_GROUPNAME_SIZE]; /* the owner's group */
 	char hostname[MAX_HOSTNAME_SIZE]; /* the requested hostname */
 	uint8_t cookie[COOKIE_SIZE]; /* the cookie associated with the session */
+
+	void * auth_ctx; /* the context of authentication */
+	unsigned auth_status; /* PS_AUTH_ */
+	unsigned auth_reqs; /* the number of requests received */
 };
 
 struct proc_list_st {
@@ -62,6 +82,10 @@ struct proc_list_st {
 	unsigned int total;
 };
 
+struct script_list_st {
+	struct list_head head;
+};
+
 struct banned_st {
 	struct list_node list;
 	time_t failed_time;	/* The time authentication failed */
@@ -77,21 +101,30 @@ typedef struct main_server_st {
 	struct cfg_st *config;
 	struct tun_st *tun;
 	hash_db_st *tls_db;
-	hash_db_st *cookie_db;
+	
+	uint8_t cookie_key[16];
 
 	/* tls credentials */
 	struct tls_st creds;
 
 	struct listen_list_st llist;
 	struct proc_list_st clist;
+	struct script_list_st script_list;
 	struct ban_list_st ban_list;
+	
+	char socket_file[_POSIX_PATH_MAX];
+	pid_t sec_mod_pid;
+	
+	unsigned active_clients;
+
+	void * auth_extra;
 } main_server_st;
 
 void clear_lists(main_server_st *s);
 
 int handle_commands(main_server_st *s, struct proc_st* cur);
 
-int user_connected(main_server_st *s, struct proc_st* cur, struct lease_st*);
+int user_connected(main_server_st *s, struct proc_st* cur);
 void user_disconnected(main_server_st *s, struct proc_st* cur);
 
 void expire_tls_sessions(main_server_st *s);
@@ -115,17 +148,22 @@ void
 __attribute__ ((format(printf, 4, 5)))
     mslog(const main_server_st * s, const struct proc_st* proc,
     	int priority, const char *fmt, ...);
+void  mslog_hex(const main_server_st * s, const struct proc_st* proc,
+    	int priority, const char *prefix, uint8_t* bin, unsigned bin_size);
 
-int open_tun(main_server_st* s, struct lease_st** l);
+int open_tun(main_server_st* s, struct proc_st* proc);
 int set_tun_mtu(main_server_st* s, struct proc_st * proc, unsigned mtu);
 
+int send_auth_reply_msg(main_server_st* s, struct proc_st* proc);
 int send_auth_reply(main_server_st* s, struct proc_st* proc,
-			cmd_auth_reply_t r, struct lease_st* lease);
+			cmd_auth_reply_t r);
 int handle_auth_cookie_req(main_server_st* s, struct proc_st* proc,
- 			   const struct cmd_auth_cookie_req_st * req, struct lease_st **lease);
-int generate_and_store_vals(main_server_st *s, struct proc_st* proc);
+ 			   const struct cmd_auth_cookie_req_st * req);
+int generate_cookie(main_server_st *s, struct proc_st* proc);
+int handle_auth_init(main_server_st *s, struct proc_st* proc,
+		     const struct cmd_auth_init_st * req);
 int handle_auth_req(main_server_st *s, struct proc_st* proc,
-		   const struct cmd_auth_req_st * req, struct lease_st **lease);
+		     struct cmd_auth_req_st * req);
 
 int check_multiple_users(main_server_st *s, struct proc_st* proc);
 
@@ -133,4 +171,8 @@ void add_to_ip_ban_list(main_server_st* s, struct sockaddr_storage *addr, sockle
 void expire_banned(main_server_st* s);
 int check_if_banned(main_server_st* s, struct sockaddr_storage *addr, socklen_t addr_len);
 
+int handle_script_exit(main_server_st *s, struct proc_st* proc, int code);
+
+void run_sec_mod(main_server_st * s);
+
 #endif
