@@ -19,7 +19,10 @@
 #include <config.h>
 #include <stdarg.h>
 #include <stdio.h>
-#include <sys/prctl.h>
+#if !defined(HAVE_SETPROCTITLE)
+
+# if defined(__linux__)
+#  include <sys/prctl.h>
 
 /* This sets the proccess title as shown in top, but not in ps (*@#%@).
  * To change the ps name in Linux, one needs to do master black magic
@@ -27,7 +30,7 @@
  */
 void setproctitle (const char *fmt, ...)
 {
-#ifdef PR_SET_NAME
+#  ifdef PR_SET_NAME
 	char name[16];
 	va_list args;
 
@@ -36,5 +39,15 @@ void setproctitle (const char *fmt, ...)
 	va_end(args);
 
 	prctl (PR_SET_NAME, name);
-#endif
+#  endif
 }
+# else /* not linux */
+
+void setproctitle (const char *fmt, ...)
+{
+
+}
+
+# endif /* __linux__ */
+
+#endif /* HAVE_SETPROCTITLE */
