@@ -21,6 +21,7 @@
 #include <gnutls/gnutls.h>
 #include <gnutls/crypto.h>
 #include <gnutls/pkcs11.h>
+#include <gnutls/abstract.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <stdarg.h>
@@ -36,7 +37,9 @@
 #include <vpn.h>
 #include <main.h>
 #include <worker.h>
-
+#include <sys/un.h>
+#include <sys/uio.h>
+#include <c-ctype.h>
 
 ssize_t tls_send(gnutls_session_t session, const void *data,
 			size_t data_size)
@@ -60,6 +63,31 @@ ssize_t tls_send(gnutls_session_t session, const void *data,
 	return data_size;
 }
 
+/* Same as tls_send() but will not retry on EAGAIN errors */
+ssize_t tls_send_nowait(gnutls_session_t session, const void *data,
+			size_t data_size)
+{
+	int ret;
+	int left = data_size;
+	const uint8_t* p = data;
+
+	while(left > 0) {
+		ret = gnutls_record_send(session, p, data_size);
+		if (ret < 0 && ret != GNUTLS_E_INTERRUPTED) {
+			if (ret == GNUTLS_E_AGAIN)
+			  return data_size;
+			return ret;
+		}
+	
+		if (ret > 0) {
+			left -= ret;
+			p += ret;
+		}
+	}
+	
+	return data_size;
+}
+
 ssize_t tls_send_file(gnutls_session_t session, const char *file)
 {
 FILE* fp;
@@ -97,13 +125,14 @@ int __attribute__ ((format(printf, 2, 3)))
 {
 	char buf[1024];
 	va_list args;
+	size_t s;
 
 	buf[1023] = 0;
 
 	va_start(args, fmt);
-	vsnprintf(buf, 1023, fmt, args);
+	s = vsnprintf(buf, 1023, fmt, args);
 	va_end(args);
-	return tls_send(session, buf, strlen(buf));
+	return tls_send(session, buf, s);
 
 }
 
@@ -189,7 +218,7 @@ static int verify_certificate_cb(gnutls_session_t session)
 
 	ws = gnutls_session_get_ptr(session);
 	if (ws == NULL) {
-		syslog(LOG_ERR, "%s:%d: could not obtain worker state.", __func__, __LINE__);
+		syslog(LOG_ERR, "%s:%d: could not obtain worker state", __func__, __LINE__);
 		return -1;
 	}
 	
@@ -208,7 +237,6 @@ static int verify_certificate_cb(gnutls_session_t session)
 	}
 
 	if (status != 0) {
-#if GNUTLS_VERSION_NUMBER > 0x030106
 		gnutls_datum_t out;
 		int type = gnutls_certificate_type_get(session);
 
@@ -221,9 +249,6 @@ static int verify_certificate_cb(gnutls_session_t session)
 		oclog(ws, LOG_INFO, "client certificate verification failed: %s", out.data);
 
 		gnutls_free(out.data);
-#else
-		oclog(ws, LOG_INFO, "client certificate verification failed.");
-#endif
 
 		goto fail;
 	} else {
@@ -241,103 +266,6 @@ fail:
 
 }
 
-int pin_callback (void *user, int attempt, const char *token_url,
-	const char *token_label, unsigned int flags, char *pin,
-	size_t pin_max)
-{
-struct tls_st * ts = user;
-int srk = 0;
-const char* p;
-unsigned len;
-
-	if (flags & GNUTLS_PIN_FINAL_TRY) {
-		syslog(LOG_ERR, "PIN callback: final try before locking; not attempting to unlock");
-		return -1;
-	}
-
-	if (flags & GNUTLS_PIN_WRONG) {
-		syslog(LOG_ERR, "PIN callback: wrong PIN was entered for '%s' (%s)", token_label, token_url);
-		return -1;
-	}
-
-	if (ts->pin[0] == 0) {
-		syslog(LOG_ERR, "PIN required for '%s' but pin-file was not set", token_label);
-		return -1;
-	}
-
-	if (strcmp(token_url, "SRK") == 0 || strcmp(token_label, "SRK") == 0) {
-		srk = 1;
-		p = ts->srk_pin;
-	} else {
-		p = ts->pin;
-	}
-
-	if (srk != 0 && ts->srk_pin[0] == 0) {
-		syslog(LOG_ERR, "PIN required for '%s' but srk-pin-file was not set", token_label);
-		return -1;
-	}
-	
-	len = strlen(p);
-	if (len > pin_max-1) {
-		syslog(LOG_ERR, "Too long PIN (%u chars)", len);
-		return -1;
-	}
-	
-	memcpy(pin, p, len);
-	pin[len] = 0;
-	
-	return 0;
-}
-
-static
-int load_pins(main_server_st* s)
-{
-int fd, ret;
-
-	s->creds.srk_pin[0] = 0;
-	s->creds.pin[0] = 0;
-
-	if (s->config->srk_pin_file != NULL) {
-		fd = open(s->config->srk_pin_file, O_RDONLY);
-		if (fd < 0) {
-			mslog(s, NULL, LOG_ERR, "could not open SRK PIN file '%s'", s->config->srk_pin_file);
-			return -1;
-		}
-	
-		ret = read(fd, s->creds.srk_pin, sizeof(s->creds.srk_pin));
-		close(fd);
-		if (ret <= 1) {
-			mslog(s, NULL, LOG_ERR, "could not read from PIN file '%s'", s->config->srk_pin_file);
-			return -1;
-		}
-	
-		if (s->creds.srk_pin[ret-1] == '\n' || s->creds.srk_pin[ret-1] == '\r')
-			s->creds.srk_pin[ret-1] = 0;
-		s->creds.srk_pin[ret] = 0;
-	}
-
-	if (s->config->pin_file != NULL) {
-		fd = open(s->config->pin_file, O_RDONLY);
-		if (fd < 0) {
-			mslog(s, NULL, LOG_ERR, "could not open PIN file '%s'", s->config->pin_file);
-			return -1;
-		}
-	
-		ret = read(fd, s->creds.pin, sizeof(s->creds.pin));
-		close(fd);
-		if (ret <= 1) {
-			mslog(s, NULL, LOG_ERR, "could not read from PIN file '%s'", s->config->pin_file);
-			return -1;
-		}
-	
-		if (s->creds.pin[ret-1] == '\n' || s->creds.pin[ret-1] == '\r')
-			s->creds.pin[ret-1] = 0;
-		s->creds.pin[ret] = 0;
-	}
-	
-	return 0;
-}
-
 void tls_global_init(main_server_st* s)
 {
 int ret;
@@ -350,6 +278,237 @@ int ret;
 	return;
 }
 
+/* Checks, if there is a single certificate specified, whether it
+ * is compatible with all ciphersuites */
+static void certificate_check(main_server_st *s)
+{
+gnutls_datum_t data = {NULL, 0};
+gnutls_x509_crt_t crt = NULL;
+int ret;
+unsigned usage;
+
+	if (s->config->cert_size > 1)
+		return;
+
+	if (gnutls_url_is_supported(s->config->cert[0]) == 0) {
+		/* no URL */
+		ret = gnutls_load_file(s->config->cert[0], &data);
+		if (ret < 0)
+			return;
+		
+		ret = gnutls_x509_crt_init(&crt);
+		GNUTLS_FATAL_ERR(ret);
+
+		ret = gnutls_x509_crt_import(crt, &data, GNUTLS_X509_FMT_PEM);
+		GNUTLS_FATAL_ERR(ret);
+		
+		ret = gnutls_x509_crt_get_pk_algorithm(crt, NULL);
+		if (ret != GNUTLS_PK_RSA)
+			goto cleanup;
+		
+		ret = gnutls_x509_crt_get_key_usage(crt, &usage, NULL);
+		if (ret >= 0) {
+			if (!(usage & GNUTLS_KEY_KEY_ENCIPHERMENT)) {
+				mslog(s, NULL, LOG_WARNING, "server certificate key usage prevents key encipherment; unable to support the RSA ciphersuites\n");
+				if (s->config->dh_params_file != NULL)
+					mslog(s, NULL, LOG_WARNING, "no DH-params file specified; server will be limited to ECDHE ciphersuites\n");
+			}
+		}
+	}
+
+cleanup:
+	if (crt != NULL)
+        	gnutls_x509_crt_deinit(crt);
+	gnutls_free(data.data);
+	return;
+}
+
+static void set_dh_params(main_server_st* s, gnutls_certificate_credentials_t cred)
+{
+gnutls_datum_t data;
+int ret;
+
+	if (s->config->dh_params_file != NULL) {
+		ret = gnutls_dh_params_init (&s->creds.dh_params);
+		GNUTLS_FATAL_ERR(ret);
+
+		ret = gnutls_load_file(s->config->dh_params_file, &data);
+		GNUTLS_FATAL_ERR(ret);
+
+		ret = gnutls_dh_params_import_pkcs3(s->creds.dh_params, &data, GNUTLS_X509_FMT_PEM);
+		GNUTLS_FATAL_ERR(ret);
+
+		gnutls_free(data.data);
+
+		gnutls_certificate_set_dh_params(cred, s->creds.dh_params);
+	}
+}
+
+struct key_cb_data {
+	unsigned idx; /* the index of the key */
+	struct sockaddr_un sa;
+	unsigned sa_len;
+};
+
+static
+int key_cb_common_func (gnutls_privkey_t key, void* userdata, const gnutls_datum_t * raw_data,
+	gnutls_datum_t * output, unsigned type)
+{
+	struct key_cb_data* cdata = userdata;
+	int sd, ret, e;
+	uint8_t header[2];
+	struct iovec iov[2];
+	uint16_t length;
+	
+	output->data = NULL;
+	
+	sd = socket(AF_UNIX, SOCK_STREAM, 0);
+	if (sd == -1) {
+		e = errno;
+		syslog(LOG_ERR, "error opening socket: %s", strerror(e));
+		return GNUTLS_E_INTERNAL_ERROR;
+	}
+	
+	ret = connect(sd, (struct sockaddr *)&cdata->sa, cdata->sa_len);
+	if (ret == -1) {
+		e = errno;
+		syslog(LOG_ERR, "error connecting to sec-mod socket '%s': %s", 
+			cdata->sa.sun_path, strerror(e));
+		return GNUTLS_E_INTERNAL_ERROR;
+	}
+	
+	header[0] = cdata->idx;
+	header[1] = type;
+	
+	iov[0].iov_base = header;
+	iov[0].iov_len = 2;
+	iov[1].iov_base = raw_data->data;
+	iov[1].iov_len = raw_data->size;
+	
+	ret = writev(sd, iov, 2);
+	if (ret == -1) {
+		e = errno;
+		syslog(LOG_ERR, "error writing to sec-mod: %s", strerror(e));
+		goto error;
+	}
+	
+	ret = recv(sd, &length, 2, 0);
+	if (ret < 2) {
+		e = errno;
+		syslog(LOG_ERR, "error reading from sec-mod: %s", strerror(e));
+		goto error;
+	}
+
+	output->size = length;
+	output->data = gnutls_malloc(output->size);
+	if (output->data == NULL) {
+		syslog(LOG_ERR, "error allocating memory");
+		goto error;
+	}
+	
+	ret = recv(sd, output->data, output->size, 0);
+	if (ret <= 0) {
+		e = errno;
+		syslog(LOG_ERR, "error reading from sec-mod: %s", strerror(e));
+		goto error;
+	}
+	
+	output->size = ret;
+	
+	close(sd);
+	return 0;
+
+error:
+	close(sd);
+	gnutls_free(output->data);
+	return GNUTLS_E_INTERNAL_ERROR;
+
+}
+
+static
+int key_cb_sign_func (gnutls_privkey_t key, void* userdata, const gnutls_datum_t * raw_data,
+	gnutls_datum_t * signature)
+{
+	return key_cb_common_func(key, userdata, raw_data, signature, 'S');
+}
+
+static int key_cb_decrypt_func(gnutls_privkey_t key, void* userdata, const gnutls_datum_t * ciphertext,
+	gnutls_datum_t * plaintext)
+{
+	return key_cb_common_func(key, userdata, ciphertext, plaintext, 'D');
+}
+
+static void key_cb_deinit_func(gnutls_privkey_t key, void* userdata)
+{
+	free(userdata);
+}
+
+static
+int load_key_files(main_server_st *s)
+{
+int ret;
+gnutls_pcert_st *pcert_list;
+unsigned pcert_list_size, i;
+gnutls_privkey_t key;
+gnutls_datum_t data;
+struct key_cb_data * cdata;
+
+	for (i=0;i<s->config->key_size;i++) {
+		/* load the certificate */
+		if (gnutls_url_is_supported(s->config->cert[i]) != 0) {
+			mslog(s, NULL, LOG_ERR, "Loading a certificate from '%s' is unsupported", s->config->cert[i]);
+			return -1;
+		} else {
+			ret = gnutls_load_file(s->config->cert[i], &data);
+			if (ret < 0) {
+				mslog(s, NULL, LOG_ERR, "error loading file '%s'", s->config->cert[i]);
+				return -1;
+			}
+		
+			pcert_list_size = 8;
+			pcert_list = gnutls_malloc(sizeof(pcert_list[0])*pcert_list_size);
+			if (pcert_list == NULL) {
+				mslog(s, NULL, LOG_ERR, "error allocating memory");
+				return -1;
+			}
+
+			ret = gnutls_pcert_list_import_x509_raw(pcert_list, &pcert_list_size,
+				&data, GNUTLS_X509_FMT_PEM, GNUTLS_X509_CRT_LIST_FAIL_IF_UNSORTED|GNUTLS_X509_CRT_LIST_IMPORT_FAIL_IF_EXCEED);
+			GNUTLS_FATAL_ERR(ret);
+			
+			gnutls_free(data.data);
+		}
+
+		ret = gnutls_privkey_init(&key);
+		GNUTLS_FATAL_ERR(ret);
+
+		cdata = malloc(sizeof(*cdata));
+		if (cdata == NULL) {
+			mslog(s, NULL, LOG_ERR, "error allocating memory");
+			return -1;
+		}
+		
+		cdata->idx = i;
+
+		memset(&cdata->sa, 0, sizeof(cdata->sa));
+		cdata->sa.sun_family = AF_UNIX;
+		snprintf(cdata->sa.sun_path, sizeof(cdata->sa.sun_path), "%s", s->socket_file);
+		cdata->sa_len = SUN_LEN(&cdata->sa);
+
+		/* load the private key */
+		ret = gnutls_privkey_import_ext2(key, gnutls_pubkey_get_pk_algorithm(pcert_list[0].pubkey, NULL),
+			cdata, key_cb_sign_func, key_cb_decrypt_func,
+			key_cb_deinit_func, GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE);
+		GNUTLS_FATAL_ERR(ret);
+
+		ret = gnutls_certificate_set_key(s->creds.xcred, NULL, 0, pcert_list,
+				pcert_list_size, key);
+		GNUTLS_FATAL_ERR(ret);
+	}
+
+	return 0;
+}
+
 /* reload key files etc. */
 void tls_global_init_certs(main_server_st* s)
 {
@@ -367,28 +526,19 @@ const char* perr;
 	ret = gnutls_certificate_allocate_credentials(&s->creds.xcred);
 	GNUTLS_FATAL_ERR(ret);
 
-	ret = load_pins(s);
-	if (ret < 0) {
+	set_dh_params(s, s->creds.xcred);
+	
+	if (s->config->key_size == 0 || s->config->cert_size == 0) {
+		mslog(s, NULL, LOG_ERR, "no certificate or key files were specified"); 
 		exit(1);
 	}
 
-	gnutls_certificate_set_pin_function (s->creds.xcred, pin_callback, &s->creds);
+	certificate_check(s);
 	
-	if (s->config->key != NULL && strncmp(s->config->key, "pkcs11:", 7) != 0) {
-		ret =
-		    gnutls_certificate_set_x509_key_file(s->creds.xcred, s->config->cert,
-						 s->config->key,
-						 GNUTLS_X509_FMT_PEM);
-		if (ret < 0) {
-			mslog(s, NULL, LOG_ERR, "error setting the certificate (%s) or key (%s) files: %s\n",
-				s->config->cert, s->config->key, gnutls_strerror(ret));
-			exit(1);
-		}
-	} else {
-#ifndef HAVE_PKCS11
-		mslog(s, NULL, LOG_ERR, "cannot load key, GnuTLS is compiled without pkcs11 support\n");
+	ret = load_key_files(s);
+	if (ret < 0) {
+		mslog(s, NULL, LOG_ERR, "error loading the certificate or key file");
 		exit(1);
-#endif	
 	}
 
 	if (s->config->cert_req != GNUTLS_CERT_IGNORE) {
@@ -398,12 +548,12 @@ const char* perr;
 								   s->config->ca,
 								   GNUTLS_X509_FMT_PEM);
 			if (ret < 0) {
-				mslog(s, NULL, LOG_ERR, "error setting the CA (%s) file.\n",
+				mslog(s, NULL, LOG_ERR, "error setting the CA (%s) file",
 					s->config->ca);
 				exit(1);
 			}
 
-			mslog(s, NULL, LOG_INFO, "processed %d CA certificate(s).\n", ret);
+			mslog(s, NULL, LOG_INFO, "processed %d CA certificate(s)", ret);
 		}
 
 		if (s->config->crl != NULL) {
@@ -412,7 +562,7 @@ const char* perr;
 								 s->config->crl,
 								 GNUTLS_X509_FMT_PEM);
 			if (ret < 0) {
-				mslog(s, NULL, LOG_ERR, "error setting the CRL (%s) file.\n",
+				mslog(s, NULL, LOG_ERR, "error setting the CRL (%s) file",
 					s->config->crl);
 				exit(1);
 			}
@@ -424,7 +574,7 @@ const char* perr;
 
 	ret = gnutls_priority_init(&s->creds.cprio, s->config->priorities, &perr);
 	if (ret == GNUTLS_E_PARSING_ERROR)
-		mslog(s, NULL, LOG_ERR, "error in TLS priority string: %s\n", perr);
+		mslog(s, NULL, LOG_ERR, "error in TLS priority string: %s", perr);
 	GNUTLS_FATAL_ERR(ret);
 	
 	if (s->config->ocsp_response != NULL) {
@@ -436,51 +586,14 @@ const char* perr;
 	return;
 }
 
-int tls_global_init_client(worker_st* ws)
-{
-#ifdef HAVE_PKCS11
-int ret;
-
-	/* when we have PKCS #11 keys we cannot open them and then fork(), we need
-	 * to open them at the process they are going to be used. */
-	if (ws->config->key != NULL && strncmp(ws->config->key, "pkcs11:", 7) == 0) {
-		ret = gnutls_pkcs11_reinit();
-		if (ret < 0) {
-			oclog(ws, LOG_ERR, "could not reinitialize PKCS #11 subsystem: %s\n",
-				gnutls_strerror(ret));
-			return -1;
-
-		}
-
-		ret =
-		    gnutls_certificate_set_x509_key_file(ws->creds->xcred, ws->config->cert,
-						 ws->config->key,
-						 GNUTLS_X509_FMT_PEM);
-		if (ret < 0) {
-			oclog(ws, LOG_ERR, "error setting the certificate (%s) or key (%s) files: %s\n",
-				ws->config->cert, ws->config->key, gnutls_strerror(ret));
-			return -1;
-		}
-	}
-#endif
-
-	return 0;
-}
-
 void tls_cork(gnutls_session_t session)
 {
-#if GNUTLS_VERSION_NUMBER > 0x030109
 	gnutls_record_cork(session);
-#endif
 }
 
 int tls_uncork(gnutls_session_t session)
 {
-#if GNUTLS_VERSION_NUMBER > 0x030109
 	return gnutls_record_uncork(session, GNUTLS_RECORD_WAIT);
-#else
-	return 0;
-#endif
 }
 
 void *calc_sha1_hash(char* file, unsigned cert)
@@ -490,6 +603,7 @@ gnutls_datum_t data;
 uint8_t digest[20];
 char * retval;
 gnutls_x509_crt_t crt;
+unsigned i;
 
 	ret = gnutls_load_file(file, &data);
 	if (ret < 0) {
@@ -516,14 +630,14 @@ gnutls_x509_crt_t crt;
 	gnutls_free(data.data);
 
 	if (ret < 0) {
-		fprintf(stderr, "Error calculating hash of '%s': %s\n", file, gnutls_strerror(ret));
+		fprintf(stderr, "error calculating hash of '%s': %s", file, gnutls_strerror(ret));
 		exit(1);
 	}
 	
 	size_t ret_size = sizeof(digest)*2+1;
 	retval = malloc(ret_size);
 	if (retval == NULL) {
-		fprintf(stderr, "Memory error\n");
+		fprintf(stderr, "memory error");
 		exit(1);
 	}
 	
@@ -531,10 +645,66 @@ gnutls_x509_crt_t crt;
 	data.size = sizeof(digest);
 	ret = gnutls_hex_encode(&data, retval, &ret_size);
 	if (ret < 0) {
-		fprintf(stderr, "Error in hex encode: %s\n", gnutls_strerror(ret));
+		fprintf(stderr, "error in hex encode: %s", gnutls_strerror(ret));
 		exit(1);
 	}
-	retval[ret_size] = 0;
+	if (retval[ret_size-1] == 0) ret_size--; /* remove the null terminator */
+	
+	/* convert to all caps */
+	for (i=0;i<ret_size;i++)
+	        retval[i] = c_toupper(retval[i]);
 	
 	return retval;
 }
+
+size_t tls_get_overhead(gnutls_protocol_t version, gnutls_cipher_algorithm_t cipher, gnutls_mac_algorithm_t mac)
+{
+unsigned iv_size, overhead = 0, t;
+unsigned block_size;
+
+	block_size = gnutls_cipher_get_block_size(cipher);
+#if GNUTLS_VERSION_NUMBER >= 0x030200
+	iv_size = gnutls_cipher_get_iv_size(cipher);
+#else
+	iv_size = block_size;
+#endif
+	
+	switch(version) {
+		case GNUTLS_DTLS0_9:
+		case GNUTLS_DTLS1_0:
+#if GNUTLS_VERSION_NUMBER >= 0x030200
+		case GNUTLS_DTLS1_2:
+#endif
+			overhead += 13;
+			break;
+		default:
+			overhead += 5;
+			break;
+	}
+	
+	switch(cipher) {
+		case GNUTLS_CIPHER_3DES_CBC:
+		case GNUTLS_CIPHER_AES_128_CBC:
+		case GNUTLS_CIPHER_AES_256_CBC:
+		case GNUTLS_CIPHER_CAMELLIA_128_CBC:
+		case GNUTLS_CIPHER_CAMELLIA_256_CBC:
+		case GNUTLS_CIPHER_AES_192_CBC:
+		case GNUTLS_CIPHER_CAMELLIA_192_CBC:
+			overhead += block_size; /* max pad */
+			overhead += iv_size; /* explicit IV */
+			break;
+		case GNUTLS_CIPHER_AES_128_GCM:
+		case GNUTLS_CIPHER_AES_256_GCM:
+			overhead += iv_size; /* explicit IV */
+			overhead += block_size; /* tag size */
+			break;
+		default:
+			break;
+	}
+
+	t = gnutls_hmac_get_len(mac);
+	if (t > 0)
+		overhead += t;
+		
+	return overhead;
+}
