@@ -14,15 +14,17 @@ int __attribute__ ((format(printf, 2, 3)))
 ssize_t tls_recv(gnutls_session_t session, void *data, size_t data_size);
 ssize_t tls_send(gnutls_session_t session, const void *data,
 			size_t data_size);
+ssize_t tls_send_nowait(gnutls_session_t session, const void *data,
+			size_t data_size);
 
 void tls_cork(gnutls_session_t session);
 int tls_uncork(gnutls_session_t session);
 
 void tls_global_init(struct main_server_st* s);
-int tls_global_init_client(struct worker_st* ws);
 void tls_global_init_certs(struct main_server_st* s);
 
 ssize_t tls_send_file(gnutls_session_t session, const char *file);
+size_t tls_get_overhead(gnutls_protocol_t, gnutls_cipher_algorithm_t, gnutls_mac_algorithm_t);
 
 #define GNUTLS_FATAL_ERR(x) \
         if (x < 0 && gnutls_error_is_fatal (x) != 0) { \
@@ -51,13 +53,10 @@ void tls_close(gnutls_session_t session);
 void tls_fatal_close(gnutls_session_t session,
 			    gnutls_alert_description_t a);
 
-#define MAX_PIN_SIZE GNUTLS_PKCS11_MAX_PIN_LEN
-
 struct tls_st {
 	gnutls_certificate_credentials_t xcred;
 	gnutls_priority_t cprio;
-	char pin[MAX_PIN_SIZE];
-	char srk_pin[MAX_PIN_SIZE];
+	gnutls_dh_params_t dh_params;
 };
 
 typedef struct
