@@ -20,9 +20,9 @@
 
 #include <fcntl.h>
 #include <sys/types.h>
+#include <sys/stat.h>
 #include <arpa/inet.h>
 #include <sys/socket.h>
-#include <linux/if_tun.h>
 #include <net/if.h>
 #include <sys/ioctl.h>
 #include <stdio.h>
@@ -30,6 +30,13 @@
 #include <stdlib.h>
 #include <errno.h>
 #include <cloexec.h>
+#include <icmp-ping.h>
+
+#ifdef __linux__
+# include <linux/if_tun.h>
+#else
+# include <net/if_tun.h>
+#endif
 
 #include <netdb.h>
 
@@ -38,34 +45,43 @@
 #include <main.h>
 #include <ccan/list/list.h>
 
-static int bignum_add1 (uint8_t * num, unsigned size)
+/* INCREMENT taken from nettle's macros */
+/* Requires that size > 0 */
+#define INCREMENT(size, ctr)			\
+  do {						\
+    unsigned increment_i = (size) - 1;		\
+    if (++(ctr)[increment_i] == 0)		\
+      while (increment_i > 0			\
+	     && ++(ctr)[--increment_i] == 0 )	\
+	;					\
+  } while (0)
+
+static void bignum_add (uint8_t * num, unsigned size, unsigned step)
 {
-  register int i, y = 0;
+  register int i;
+  register unsigned tmp, y;
 
   for (i = size-1; i >= 0; i--)
     {
-      y = 0;
-      if (num[i] == 0xff)
-        {
-          num[i] = 0;
-          y = 1;
-        }
+      tmp = num[i];
+      
+      num[i] += step;
+      if (num[i] < tmp)
+        y = 1;
       else
-        num[i]++;
+        y = 0;
 
       if (y == 0)
         break;
     }
-
-  return 0;
 }
 
-static int get_avail_network_addresses(const struct cfg_st *config, const struct lease_st *last4,
+static int get_avail_network_addresses(main_server_st* s, const struct lease_st *last4,
 					const struct lease_st *last6, struct lease_st* lease)
 {
 	struct sockaddr_storage tmp, mask, network;
 	unsigned i;
-	int ret;
+	int ret, step;
 
 	lease->rip4_len = 0;
 	lease->lip4_len = 0;
@@ -74,138 +90,116 @@ static int get_avail_network_addresses(const struct cfg_st *config, const struct
 	
 	memset(&tmp, 0, sizeof(tmp));
 
-	if (config->network.ipv4 && config->network.ipv4_netmask) {
+	if (s->config->network.ipv4 && s->config->network.ipv4_netmask) {
 		ret =
-		    inet_pton(AF_INET, config->network.ipv4, SA_IN_P(&network));
+		    inet_pton(AF_INET, s->config->network.ipv4, SA_IN_P(&network));
 
 		if (ret != 1) {
-			syslog(LOG_ERR, "Error reading IP: %s\n", config->network.ipv4);
+			mslog(s, NULL, LOG_ERR, "Error reading IP: %s\n", s->config->network.ipv4);
 			return -1;
 		}
 
 		ret =
-		    inet_pton(AF_INET, config->network.ipv4_netmask, SA_IN_P(&mask));
+		    inet_pton(AF_INET, s->config->network.ipv4_netmask, SA_IN_P(&mask));
 	
 		if (ret != 1) {
-			syslog(LOG_ERR, "Error reading mask: %s\n", config->network.ipv4_netmask);
+			mslog(s, NULL, LOG_ERR, "Error reading mask: %s\n", s->config->network.ipv4_netmask);
 			return -1;
 		}
 
 		/* mask the network (just in case it is wrong) */
 		for (i=0;i<sizeof(struct in_addr);i++)
 			SA_IN_U8_P(&network)[i] &= (SA_IN_U8_P(&mask)[i]);
+        	((struct sockaddr_in*)&network)->sin_family = AF_INET;
 
-		memcpy(&tmp, &network, sizeof(tmp));
-		((struct sockaddr_in*)&tmp)->sin_family = AF_INET;
-		((struct sockaddr_in*)&network)->sin_family = AF_INET;
-	}
-	
+        	if (last4 != NULL) {
+        	        memcpy(&tmp, &last4->rip4, last4->rip4_len);
+                } else {
+        		memcpy(&tmp, &network, sizeof(tmp));
+        		((struct sockaddr_in*)&tmp)->sin_family = AF_INET;
+                }
 
-	/* read the network */
-	if (last4 == NULL && (config->network.ipv4 && config->network.ipv4_netmask)) {
-	
-		/* add one to get local IP */
-		i = sizeof(struct in_addr)-1;
-		SA_IN_U8_P(&tmp)[i]++;
-	
 		lease->lip4_len = sizeof(struct sockaddr_in);
-		memcpy(&lease->lip4, &tmp, lease->lip4_len);
-
-		/* add one to get remote IP */
-		i = sizeof(struct in_addr)-1;
-		SA_IN_U8_P(&tmp)[i]++;
-
-		lease->rip4_len = sizeof(struct sockaddr_in);
-		memcpy(&lease->rip4, &tmp, lease->rip4_len);
-	} else if (last4 != NULL) {
-
-		lease->lip4_len = last4->rip4_len;
-		memcpy(&lease->lip4, &last4->rip4, last4->rip4_len);
-
-		bignum_add1(SA_IN_U8_P(&lease->lip4), sizeof(struct in_addr));
-		if (SA_IN_U8_P(&lease->lip4)[3] == 255) /* broadcast */
-			bignum_add1(SA_IN_U8_P(&lease->lip4), sizeof(struct in_addr));
-
-		lease->rip4_len = last4->rip4_len;
-		memcpy(&lease->rip4, &lease->lip4, lease->rip4_len);
-		bignum_add1(SA_IN_U8_P(&lease->rip4), sizeof(struct in_addr));
-
-		/* mask the last IP with the netmask */
-		memcpy(&tmp, &lease->rip4, lease->rip4_len);
-		for (i=0;i<sizeof(struct in_addr);i++)
-			SA_IN_U8_P(&tmp)[i] &= (SA_IN_U8_P(&mask)[i]);
+		memcpy(&lease->lip4, &tmp, sizeof(struct sockaddr_in));
 		
-		/* the result should match the network */
-		if (memcmp(SA_IN_U8_P(&network), SA_IN_U8_P(&tmp), sizeof(struct in_addr)) != 0) {
-			syslog(LOG_ERR, "Reached limit of maximum (v4) IPs.\n");
-			return -1;
-		}
+		step = 1;
+		do {
+        		bignum_add(SA_IN_U8_P(&lease->lip4), sizeof(struct in_addr), step);
+        		if (SA_IN_U8_P(&lease->lip4)[3] == 255) /* broadcast */
+	        		bignum_add(SA_IN_U8_P(&lease->lip4), sizeof(struct in_addr), step);
+
+        		lease->rip4_len = sizeof(struct sockaddr_in);
+        		memcpy(&lease->rip4, &lease->lip4, sizeof(struct sockaddr_in));
+        		bignum_add(SA_IN_U8_P(&lease->rip4), sizeof(struct in_addr), step);
+
+        		/* mask the last IP with the netmask */
+        		memcpy(&tmp, &lease->rip4, lease->rip4_len);
+        		for (i=0;i<sizeof(struct in_addr);i++)
+        			SA_IN_U8_P(&tmp)[i] &= (SA_IN_U8_P(&mask)[i]);
+		
+        		/* the result should match the network */
+        		if (memcmp(SA_IN_U8_P(&network), SA_IN_U8_P(&tmp), sizeof(struct in_addr)) != 0) {
+        			mslog(s, NULL, LOG_ERR, "Reached limit of maximum (v4) IPs.\n");
+        			return -1;
+        		}
+                } while((step=icmp_ping4(s, (void*)&lease->lip4, (void*)&lease->rip4)) != 0);
 	}
 
-	if (config->network.ipv6 && config->network.ipv6_netmask) {
+	if (s->config->network.ipv6 && s->config->network.ipv6_netmask) {
 		ret =
-		    inet_pton(AF_INET6, config->network.ipv6, SA_IN6_P(&network));
+		    inet_pton(AF_INET6, s->config->network.ipv6, SA_IN6_P(&network));
 
 		if (ret != 1) {
-			syslog(LOG_ERR, "Error reading IP: %s\n", config->network.ipv6);
+			mslog(s, NULL, LOG_ERR, "Error reading IP: %s\n", s->config->network.ipv6);
 			return -1;
 		}
 
 		ret =
-		    inet_pton(AF_INET6, config->network.ipv6_netmask, SA_IN6_P(&mask));
+		    inet_pton(AF_INET6, s->config->network.ipv6_netmask, SA_IN6_P(&mask));
 	
 		if (ret != 1) {
-			syslog(LOG_ERR, "Error reading mask: %s\n", config->network.ipv6_netmask);
+			mslog(s, NULL, LOG_ERR, "Error reading mask: %s\n", s->config->network.ipv6_netmask);
 			return -1;
 		}
 	
 		/* mask the network */
 		for (i=0;i<sizeof(struct in6_addr);i++)
 			SA_IN6_U8_P(&network)[i] &= (SA_IN6_U8_P(&mask)[i]);
-	
-		memcpy(&tmp, &network, sizeof(tmp));
-		((struct sockaddr_in6*)&tmp)->sin6_family = AF_INET6;
 		((struct sockaddr_in6*)&network)->sin6_family = AF_INET6;
-	}
 
-	if (last6 == NULL && (config->network.ipv6 && config->network.ipv6_netmask)) {
-		/* add one to get local IP */
-		i = sizeof(struct in6_addr)-1;
-		SA_IN6_U8_P(&tmp)[i]++;
-	
+        	if (last6 != NULL) {
+        	        memcpy(&tmp, &last6->rip6, last6->rip6_len);
+                } else {
+        		memcpy(&tmp, &network, sizeof(tmp));
+	        	((struct sockaddr_in6*)&tmp)->sin6_family = AF_INET6;
+                }
+
 		lease->lip6_len = sizeof(struct sockaddr_in6);
-		memcpy(&lease->lip6, &tmp, lease->lip6_len);
+		memcpy(&lease->lip6, &tmp, sizeof(struct sockaddr_in6));
 
-		/* add one to get remote IP */
-		i = sizeof(struct in6_addr)-1;
-		SA_IN6_U8_P(&tmp)[i]++;
-	
-		lease->rip6_len = sizeof(struct sockaddr_in6);
-		memcpy(&lease->rip6, &tmp, lease->rip6_len);
-	} else if (last6 != NULL) {
+		step = 1;
+		do {
+        		bignum_add(SA_IN6_U8_P(&lease->lip6), sizeof(struct in6_addr), step);
 
-		lease->lip6_len = last6->rip6_len;
-		memcpy(&lease->lip6, &last6->rip6, last6->rip6_len);
-		bignum_add1(SA_IN6_U8_P(&lease->lip6), sizeof(struct in6_addr));
+	        	lease->rip6_len = last6->rip6_len;
+        		memcpy(&lease->rip6, &lease->lip6, lease->rip6_len);
+	        	bignum_add(SA_IN6_U8_P(&lease->rip6), sizeof(struct in6_addr), step);
 
-		lease->rip6_len = last6->rip6_len;
-		memcpy(&lease->rip6, &lease->lip6, lease->rip6_len);
-		bignum_add1(SA_IN6_U8_P(&lease->rip6), sizeof(struct in6_addr));
+        		/* mask the last IP with the netmask */
+        		memcpy(&tmp, &lease->rip6, lease->rip6_len);
+        		for (i=0;i<sizeof(struct in6_addr);i++)
+	        		SA_IN6_U8_P(&tmp)[i] &= (SA_IN6_U8_P(&mask)[i]);
 
-		/* mask the last IP with the netmask */
-		memcpy(&tmp, &lease->rip6, lease->rip6_len);
-		for (i=0;i<sizeof(struct in6_addr);i++)
-			SA_IN6_U8_P(&tmp)[i] &= (SA_IN6_U8_P(&mask)[i]);
-
-		/* the result should match the network */
-		if (memcmp(SA_IN6_U8_P(&network), SA_IN6_U8_P(&tmp), sizeof(struct in6_addr)) != 0) {
-			syslog(LOG_ERR, "Reached limit of maximum (v6) IPs.\n");
-			return -1;
-		}
+        		/* the result should match the network */
+	        	if (memcmp(SA_IN6_U8_P(&network), SA_IN6_U8_P(&tmp), sizeof(struct in6_addr)) != 0) {
+		        	mslog(s, NULL, LOG_ERR, "Reached limit of maximum (v6) IPs.\n");
+        			return -1;
+	        	}
+                } while((step=icmp_ping6(s, (void*)&lease->lip6, (void*)&lease->rip6)) != 0);
 	}
 
 	if (lease->lip6_len == 0 && lease->lip4_len == 0) {
-		syslog(LOG_ERR, "No IPv4 or IPv6 addresses are configured. Cannot obtain lease.\n");
+		mslog(s, NULL, LOG_ERR, "No IPv4 or IPv6 addresses are configured. Cannot obtain lease.\n");
 		return -1;
 	}
 
@@ -276,8 +270,8 @@ static int set_network_info( main_server_st* s, const struct lease_st *lease)
 	if (lease->lip6_len == 0 && lease->lip4_len == 0) {
 		mslog(s, NULL, LOG_ERR, "%s: Could not set any IP.\n", lease->name);
 		ret = -1;
-	} else
-		ret = 0;
+		goto cleanup;
+	}
 		
 	/* bring interface up */
 	memset(&ifr, 0, sizeof(ifr));
@@ -288,14 +282,16 @@ static int set_network_info( main_server_st* s, const struct lease_st *lease)
 	ret = ioctl(fd, SIOCSIFFLAGS, &ifr);
 	if (ret != 0) {
 		mslog(s, NULL, LOG_ERR, "%s: Could not bring up interface.\n", lease->name);
+		ret = -1;
 	}
 
+cleanup:
 	close(fd);
 	return ret;
 }
 
 
-int open_tun(main_server_st* s, struct lease_st** l)
+int open_tun(main_server_st* s, struct proc_st* proc)
 {
 	int tunfd, ret, e;
 	struct ifreq ifr;
@@ -304,15 +300,21 @@ int open_tun(main_server_st* s, struct lease_st** l)
 	struct lease_st *last4, *tmp;
 	struct lease_st *last6;
 	unsigned current = s->tun->total;
+	time_t now = time(0);
 	
 	last4 = last6 = NULL;
 		
 	/* try to re-use an address */
         list_for_each(&s->tun->head, tmp, list) {
-		if (tmp->in_use == 0) {
-			lease = tmp;
-			mslog(s, NULL, LOG_INFO, "reusing tun device %s\n", lease->name);
-			break;
+		/* if the device isn't in use by the server and the IPs
+		 *  are free. */
+		if (tmp->in_use == 0 && (tmp->available_at < now || (proc->username[0] != 0 && strcmp(proc->username, tmp->username) == 0))) {
+                        if ((tmp->lip6_len != 0 && icmp_ping6(s, (void*)&tmp->lip6, (void*)&tmp->rip6) == 0) ||
+                          (tmp->lip4_len != 0 && icmp_ping4(s, (void*)&tmp->lip4, (void*)&tmp->rip4) == 0)) {
+        			lease = tmp;
+	        		mslog(s, NULL, LOG_INFO, "reusing tun device %s\n", lease->name);
+		        	break;
+                        }
 		}
 	}
 		
@@ -333,7 +335,7 @@ int open_tun(main_server_st* s, struct lease_st** l)
 				break;
 		}
 
-		ret = get_avail_network_addresses(s->config, last4, last6, lease);
+		ret = get_avail_network_addresses(s, last4, last6, lease);
 		if (ret < 0) {
 			free(lease);
 			return -1;
@@ -342,11 +344,15 @@ int open_tun(main_server_st* s, struct lease_st** l)
 		/* Add into the list */
 		list_add_tail( &s->tun->head, &lease->list);
 		snprintf(lease->name, sizeof(lease->name), "%s%u", s->config->network.name, current);
+		snprintf(lease->username, sizeof(lease->username), "%s", proc->username);
 		s->tun->total++;
 	}
 
 	/* No need to free the lease after this point.
 	 */
+	 
+	/* Obtain a free tun device */
+#ifdef __linux__
 	tunfd = open("/dev/net/tun", O_RDWR);
 	if (tunfd < 0) {
 		int e = errno;
@@ -387,7 +393,7 @@ int open_tun(main_server_st* s, struct lease_st** l)
 		}
 	}
 
-#ifdef TUNSETGROUP
+# ifdef TUNSETGROUP
 	if (s->config->gid != -1) {
 		t = s->config->uid;
 		ret = ioctl(tunfd, TUNSETGROUP, t);
@@ -398,6 +404,31 @@ int open_tun(main_server_st* s, struct lease_st** l)
 			goto fail;
 		}
 	}
+# endif
+#else /* freebsd */
+	tunfd = open("/dev/tun", O_RDWR);
+	if (tunfd < 0) {
+		int e = errno;
+		mslog(s, NULL, LOG_ERR, "Can't open /dev/tun: %s\n",
+		       strerror(e));
+		return -1;
+	}
+	
+	/* find device name */
+	{
+		struct stat st;
+		
+		ret = fstat(tunfd, &st);
+		if (ret < 0) {
+			e = errno;
+			mslog(s, NULL, LOG_ERR, "%s: stat: %s\n", strerror(e));
+			goto fail;
+		}
+		
+		snprintf(lease->name, sizeof(lease->name), "%s", devname(st.st_rdev, S_IFCHR));
+	}
+	
+	set_cloexec_flag (tunfd, 1);
 #endif
 
 	/* set IP/mask */
@@ -406,8 +437,10 @@ int open_tun(main_server_st* s, struct lease_st** l)
 		goto fail;
 	}
 	
+	lease->in_use = 1;
+	lease->available_at = now + s->config->cookie_validity;
 	lease->fd = tunfd;
-	*l = lease;
+	proc->lease = lease;
 
 	return 0;
 fail:
