@@ -9,8 +9,10 @@ struct lease_st {
 	struct list_node list;
 
 	char name[IFNAMSIZ];
+	char username[MAX_USERNAME_SIZE]; /* owner */
 	unsigned int tun_nr;
 	unsigned int in_use;
+	time_t available_at; /* when it will be available */
 
         struct sockaddr_storage rip4;
         socklen_t rip4_len;
