@@ -10,6 +10,20 @@
 #include <sys/socket.h>
 #include <unistd.h>
 #include <net/if.h>
+#include <netinet/in.h>
+#include <minmax.h>
+
+#ifdef __GNUC__
+# define _OCSERV_GCC_VERSION (__GNUC__ * 10000 + __GNUC_MINOR__ * 100 + __GNUC_PATCHLEVEL__)
+# if _OCSERV_GCC_VERSION >= 30000
+#  define _ATTR_PACKED __attribute__ ((__packed__))
+# endif
+#endif /* __GNUC__ */
+
+#ifndef _ATTR_PACKED
+# define _ATTR_PACKED
+#endif
+
 
 #define AC_PKT_DATA             0	/* Uncompressed data */
 #define AC_PKT_DPD_OUT          3	/* Dead Peer Detection */
@@ -21,13 +35,18 @@
 
 extern int syslog_open;
 
-#define MAX(x,y) ((x)>(y)?(x):(y))
-#define MIN(x,y) ((x)<(y)?(x):(y))
-
 /* the first is generic, for the methods that require a username password */
 #define AUTH_TYPE_USERNAME_PASS (1<<0)
 #define AUTH_TYPE_PAM (1<<1 | AUTH_TYPE_USERNAME_PASS)
 #define AUTH_TYPE_CERTIFICATE (1<<2)
+#define AUTH_TYPE_PLAIN (1<<3 | AUTH_TYPE_USERNAME_PASS)
+
+#define ERR_SUCCESS 0
+#define ERR_BAD_COMMAND -2
+#define ERR_AUTH_FAIL -3
+#define ERR_AUTH_CONTINUE -4
+#define ERR_WAIT_FOR_SCRIPT -5
+#define ERR_MEM -6
 
 typedef struct 
 {
@@ -35,7 +54,6 @@ typedef struct
 	unsigned int entries;
 } hash_db_st;
 
-
 struct vpn_st {
 	char *name;	/* device name */
 	char *ipv4_netmask;
@@ -59,19 +77,26 @@ struct cfg_st {
 	unsigned int udp_port;
 	char *pin_file;
 	char *srk_pin_file;
-	char *cert;
+	char **cert;
+	unsigned cert_size;
+	char **key;
+	unsigned key_size;
 
-	char *key;
 	char *ca;
 	char *crl;
+	char *dh_params_file;
 	char *cert_user_oid;	/* The OID that will be used to extract the username */
 	char *cert_group_oid;	/* The OID that will be used to extract the groupname */
 	unsigned int auth_types;	/* or'ed sequence of AUTH_TYPE */
+	char *plain_passwd;	/* the password file in plain auth mode */
 	gnutls_certificate_request_t cert_req;
 	char *priorities;
 	char *chroot_dir;	/* where the xml files are served from */
 	char *banner;
 	char *ocsp_response; /* file with the OCSP response */
+	char *default_domain; /* domain to be advertised */
+	
+	char* socket_file_prefix;
 	time_t cookie_validity;	/* in seconds */
 	time_t min_reauth_time;	/* after a failed auth, how soon one can reauthenticate -> in seconds */
 	unsigned auth_timeout; /* timeout of HTTP auth */
@@ -86,9 +111,10 @@ struct cfg_st {
 	unsigned try_mtu; /* MTU discovery enabled */
 	unsigned force_cert_auth; /* always require client certificate */
 	unsigned rate_limit_ms; /* if non zero force a connection every rate_limit milliseconds */
-
-	/* if gdbm is there */
-	char* cookie_db_name;
+	unsigned ping_leases; /* non zero if we need to ping prior to leasing */
+	
+	unsigned output_buffer;
+	unsigned default_mtu;
 
 	char *connect_script;
 	char *disconnect_script;
@@ -101,7 +127,7 @@ struct cfg_st {
 
 	uid_t uid;
 	gid_t gid;
-
+	
 	/* the tun network */
 	struct vpn_st network;
 };
@@ -110,17 +136,16 @@ struct cfg_st {
 struct worker_st;
 struct main_server_st;
 
-#include <tun.h>
-
 #define MAX_BANNER_SIZE 256
 #define MAX_USERNAME_SIZE 64
 #define MAX_PASSWORD_SIZE 64
 #define TLS_MASTER_SIZE 48
 #define MAX_HOSTNAME_SIZE MAX_USERNAME_SIZE
 #define MAX_GROUPNAME_SIZE MAX_USERNAME_SIZE
-#define COOKIE_SIZE 32
 #define MAX_SESSION_DATA_SIZE (4*1024)
 
+#include <tun.h>
+
 const char *human_addr(const struct sockaddr *sa, socklen_t salen,
 		       void *buf, size_t buflen);
 
