@@ -18,6 +18,7 @@
 
 #include <config.h>
 #include <worker.h>
+#include <sys/ioctl.h>
 
 #ifdef HAVE_LIBSECCOMP
 
@@ -26,51 +27,63 @@
 
 int disable_system_calls(struct worker_st *ws)
 {
-	int ret, e;
+	int ret;
 	scmp_filter_ctx ctx;
-	
+
 	ctx = seccomp_init(SCMP_ACT_KILL);
 	if (ctx == NULL) {
-		oclog(ws, LOG_WARNING, "could not initialize seccomp");
+		oclog(ws, LOG_DEBUG, "could not initialize seccomp");
 		return -1;
 	}
 
-#define ADD_SYSCALL(name) \
-	ret = seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(name), 0); \
-	if (ret < 0) { \
-		e = errno; \
-		oclog(ws, LOG_WARNING, "could not add " #name " to seccomp filter: %s", strerror(e)); \
+#define ADD_SYSCALL(name, ...) \
+	ret = seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(name), __VA_ARGS__); \
+	/* libseccomp returns EDOM for pseudo-syscalls due to a bug */ \
+	if (ret < 0 && ret != -EDOM) { \
+		oclog(ws, LOG_DEBUG, "could not add " #name " to seccomp filter: %s", strerror(-ret)); \
 		ret = -1; \
 		goto fail; \
 	}
-	
-	ADD_SYSCALL(time);
-	ADD_SYSCALL(recvmsg);
-	ADD_SYSCALL(sendmsg);
-	ADD_SYSCALL(read);
-	ADD_SYSCALL(write);
-	ADD_SYSCALL(select);
-	ADD_SYSCALL(alarm);
-	ADD_SYSCALL(close);
-	ADD_SYSCALL(exit);
-	ADD_SYSCALL(exit_group);
-	ADD_SYSCALL(send);
-	ADD_SYSCALL(recv);
 
-	/* these two we need to get the MTU from
+	ADD_SYSCALL(time, 0);
+	ADD_SYSCALL(gettimeofday, 0);
+	ADD_SYSCALL(recvmsg, 0);
+	ADD_SYSCALL(sendmsg, 0);
+	ADD_SYSCALL(read, 0);
+	ADD_SYSCALL(write, 0);
+	ADD_SYSCALL(writev, 0);
+	ADD_SYSCALL(send, 0);
+	ADD_SYSCALL(recv, 0);
+
+	/* it seems we need to add sendto and recvfrom
+	 * since send() and recv() aren't real system
+	 * calls.
+	 */
+	ADD_SYSCALL(sendto, 0);
+	ADD_SYSCALL(recvfrom, 0);
+	ADD_SYSCALL(select, 0);
+	ADD_SYSCALL(alarm, 0);
+	ADD_SYSCALL(close, 0);
+	ADD_SYSCALL(exit, 0);
+	ADD_SYSCALL(exit_group, 0);
+	ADD_SYSCALL(socket, 0);
+	ADD_SYSCALL(connect, 0);
+
+	/* this we need to get the MTU from
 	 * the TUN device */
-	ADD_SYSCALL(ioctl);
-	ADD_SYSCALL(socket);
+	ADD_SYSCALL(ioctl, 1, SCMP_A1(SCMP_CMP_EQ, (int)SIOCGIFDSTADDR));
+	ADD_SYSCALL(ioctl, 1, SCMP_A1(SCMP_CMP_EQ, (int)SIOCGIFADDR));
+	ADD_SYSCALL(ioctl, 1, SCMP_A1(SCMP_CMP_EQ, (int)SIOCGIFMTU));
 
 	ret = seccomp_load(ctx);
 	if (ret < 0) {
-		oclog(ws, LOG_ERR, "could not load seccomp filter");
+		oclog(ws, LOG_DEBUG, "could not load seccomp filter");
 		ret = -1;
 		goto fail;
 	}
 	
-	seccomp_release(ctx);
 	ret = 0;
+
 fail:
 	seccomp_release(ctx);
 	return ret;
