@@ -36,9 +36,11 @@
 #include <netinet/in.h>
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <netinet/tcp.h>
 #include <arpa/inet.h>
-#include <signal.h>
+#include <system.h>
 #include <time.h>
+#include <common.h>
 
 #include <vpn.h>
 #include "ipc.h"
@@ -52,18 +54,19 @@
  * TCP (if activity occurs there).
  */
 #define UDP_SWITCH_TIME 15
+#define PERIODIC_CHECK_TIME 30
 
 /* The number of DPD packets a client skips before he's kicked */
 #define DPD_TRIES 2
 #define DPD_MAX_TRIES 3
 
 /* HTTP requests prior to disconnection */
-#define MAX_HTTP_REQUESTS 8
+#define MAX_HTTP_REQUESTS 16
 
 static int terminate = 0;
-static int parse_cstp_data(struct worker_st* ws, uint8_t* buf, size_t buf_size);
+static int parse_cstp_data(struct worker_st* ws, uint8_t* buf, size_t buf_size, time_t);
 static int parse_dtls_data(struct worker_st* ws, 
-				uint8_t* buf, size_t buf_size);
+				uint8_t* buf, size_t buf_size, time_t);
 
 static void handle_alarm(int signo)
 {
@@ -82,133 +85,52 @@ typedef int (*url_handler_fn)(worker_st*, unsigned http_ver);
 struct known_urls_st {
 	const char* url;
 	unsigned url_size;
+	unsigned partial_match;
 	url_handler_fn get_handler;
 	url_handler_fn post_handler;
 };
 
-#ifdef ANYCONNECT_CLIENT_COMPAT
-int get_config_handler(worker_st *ws, unsigned http_ver)
-{
-int ret;
-struct stat st;
 
-	oclog(ws, LOG_DEBUG, "requested config: %s", ws->req.url); 
-	if (ws->config->xml_config_file == NULL) {
-		oclog(ws, LOG_INFO, "requested config but no config file is set");
-		tls_printf(ws->session, "HTTP/1.%u 404 Not found\r\n", http_ver);
-		return -1;
-	}
-	
-	ret = stat( ws->config->xml_config_file, &st);
-	if (ret == -1) {
-		oclog(ws, LOG_INFO, "cannot load config file '%s'", ws->config->xml_config_file);
-		tls_printf(ws->session, "HTTP/1.%u 404 Not found\r\n", http_ver);
-		return -1;
-	}
-
-	tls_cork(ws->session);
-	ret = tls_printf(ws->session, "HTTP/1.%u 200 OK\r\n", http_ver);
-	if (ret < 0)
-		return -1;
-
-	ret = tls_puts(ws->session, "Connection: Keep-Alive\r\n");
-	if (ret < 0)
-		return -1;
-
-	ret = tls_puts(ws->session, "Content-Type: text/xml\r\n");
-	if (ret < 0)
-		return -1;
-
-	ret = tls_puts(ws->session, "X-Transcend-Version: 1\r\n");
-	if (ret < 0)
-		return -1;
-
-	ret = tls_printf(ws->session, "Content-Length: %u\r\n", (unsigned)st.st_size);
-	if (ret < 0)
-		return -1;
-
-	ret = tls_puts(ws->session, "\r\n");
-	if (ret < 0)
-		return -1;
-
-	ret = tls_uncork(ws->session);
-	if (ret < 0)
-		return -1;
-	
-	ret = tls_send_file(ws->session, ws->config->xml_config_file);
-	if (ret < 0) {
-		oclog(ws, LOG_ERR, "error sending file '%s': %s", ws->config->xml_config_file, gnutls_strerror(ret));
-		return -1;
-	}
-
-	return 0;
-}
-
-int get_cscot_handler(worker_st *ws, unsigned http_ver)
-{
-int ret;
-
-	oclog(ws, LOG_DEBUG, "requested CSCOT: %s", ws->req.url); 
-
-	tls_cork(ws->session);
-	ret = tls_printf(ws->session, "HTTP/1.%u 200 OK\r\n", http_ver);
-	if (ret < 0)
-		return -1;
-
-	ret = tls_puts(ws->session, "Connection: Keep-Alive\r\n");
-	if (ret < 0)
-		return -1;
-
-	ret = tls_puts(ws->session, "Content-Type: text/xml\r\n");
-	if (ret < 0)
-		return -1;
-
-	ret = tls_puts(ws->session, "X-Transcend-Version: 1\r\n");
-	if (ret < 0)
-		return -1;
-
-#define MANIFEST "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<vpn rev=\"1.0\">\n" \
-		"</vpn>\n"
-	ret = tls_printf(ws->session, "Content-Length: %u\r\n\r\n", (unsigned)sizeof(MANIFEST)-1);
-	if (ret < 0)
-		return -1;
-		
-	ret = tls_puts(ws->session, MANIFEST);
-	if (ret < 0)
-		return -1;
-
-	ret = tls_uncork(ws->session);
-	if (ret < 0)
-		return -1;
-	
-	return 0;
-}
-
-#endif
-
-#define LL(x,y,z) {x, sizeof(x)-1, y, z}
-struct known_urls_st known_urls[] = {
+#define LL(x,y,z) {x, sizeof(x)-1, 0, y, z}
+#define LL_DIR(x,y,z) {x, sizeof(x)-1, 1, y, z}
+const static struct known_urls_st known_urls[] = {
 		LL("/", get_auth_handler, post_auth_handler),
 		LL("/auth", get_auth_handler, post_auth_handler),
 #ifdef ANYCONNECT_CLIENT_COMPAT
+		LL("/2/index.html", get_empty_handler, NULL),
+		LL("/2/Linux", get_empty_handler, NULL),
+		LL("/2/Linux_64", get_empty_handler, NULL),
+		LL("/2/Windows", get_empty_handler, NULL),
+		LL("/2/binaries/vpndownloader.sh", get_dl_handler, NULL),
+		LL("/2/VPNManifest.xml", get_string_handler, NULL),
+		LL("/2/binaries/update.txt", get_string_handler, NULL),
+
+		LL("/1/index.html", get_empty_handler, NULL),
+		LL("/1/Linux", get_empty_handler, NULL),
+		LL("/1/Linux_64", get_empty_handler, NULL),
+		LL("/1/Windows", get_empty_handler, NULL),
+		LL("/1/binaries/vpndownloader.sh", get_dl_handler, NULL),
+		LL("/1/VPNManifest.xml", get_string_handler, NULL),
+		LL("/1/binaries/update.txt", get_string_handler, NULL),
 		LL("/profiles", get_config_handler, NULL),
-		LL("/+CSCOT+/translation-table", get_cscot_handler, NULL),
+		LL("/+CSCOT+/translation-table", get_string_handler, NULL),
 #endif
-		{NULL, 0, NULL, NULL}
+		{NULL, 0, 0, NULL, NULL}
 };
 
 static url_handler_fn get_url_handler(const char* url)
 {
-struct known_urls_st *p;
+const struct known_urls_st *p;
 unsigned len = strlen(url);
 
 	p = known_urls;
 	do {
-		if (p->url != NULL && (
-		        (len == p->url_size && strcmp(p->url, url)==0) ||
-			(len >= p->url_size && strncmp(p->url, url, p->url_size)==0 && 
-				(url[p->url_size] == '/' || url[p->url_size] == '?'))))
-			return p->get_handler;
+		if (p->url != NULL) {
+		        if ((len == p->url_size && strcmp(p->url, url)==0) ||
+				(len >= p->url_size && strncmp(p->url, url, p->url_size)==0 && 
+				(p->partial_match != 0 || url[p->url_size] == '/' || url[p->url_size] == '?')))
+				return p->get_handler;
+		}
 		p++;
 	} while(p->url != NULL);
 	
@@ -217,7 +139,7 @@ unsigned len = strlen(url);
 
 static url_handler_fn post_url_handler(const char* url)
 {
-struct known_urls_st *p;
+const struct known_urls_st *p;
 
 	p = known_urls;
 	do {
@@ -247,105 +169,190 @@ int url_cb(http_parser* parser, const char *at, size_t length)
 
 #define STR_HDR_COOKIE "Cookie"
 #define STR_HDR_MS "X-DTLS-Master-Secret"
+#define STR_HDR_CS "X-DTLS-CipherSuite"
 #define STR_HDR_DMTU "X-DTLS-MTU"
 #define STR_HDR_CMTU "X-CSTP-MTU"
+#define STR_HDR_ATYPE "X-CSTP-Address-Type"
 #define STR_HDR_HOST "X-CSTP-Hostname"
+
+static void value_check(struct worker_st *ws, struct http_req_st *req)
+{
+unsigned length;
+size_t nlen;
+uint8_t* p;
+char * token;
+char * str;
+
+	if (req->value.length <= 0)
+		return;
+
+	oclog(ws, LOG_DEBUG, "HTTP: %.*s: %.*s", (int)req->header.length, req->header.data,  
+		(int)req->value.length, req->value.data);
+
+	switch (req->next_header) {
+		case HEADER_MASTER_SECRET:
+			if (req->value.length < TLS_MASTER_SIZE*2) {
+				req->master_secret_set = 0;
+				return;
+			}
+			
+			length = TLS_MASTER_SIZE*2;
+
+			nlen = sizeof(req->master_secret);
+			gnutls_hex2bin((void*)req->value.data, length, req->master_secret, &nlen);
+
+			req->master_secret_set = 1;
+			break;
+		case HEADER_HOSTNAME:
+			if (req->value.length+1 > MAX_HOSTNAME_SIZE) {
+				req->hostname[0] = 0;
+				return;
+			}
+			memcpy(req->hostname, req->value.data, req->value.length);
+			req->hostname[req->value.length] = 0;
+			break;
+
+		case HEADER_DTLS_CIPHERSUITE:
+			str = (char*)req->value.data;
+			while ((token = strtok(str, ":")) != NULL) {
+#if GNUTLS_VERSION_NUMBER >= 0x030201
+				if (strcmp(token, "X-ESTREAM-SALSA20-UMAC96") == 0) {
+				        req->selected_ciphersuite = "X-ESTREAM-SALSA20-UMAC96";
+				        req->gnutls_ciphersuite = "NONE:+VERS-DTLS0.9:+COMP-NULL:+ESTREAM-SALSA20-256:+UMAC-96:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION";
+				        req->gnutls_cipher = GNUTLS_CIPHER_ESTREAM_SALSA20_256;
+				        req->gnutls_mac = GNUTLS_MAC_UMAC_96;
+				        break;
+				} else if (strcmp(token, "X-SALSA20-UMAC96") == 0) {
+				        req->gnutls_ciphersuite = "NONE:+VERS-DTLS0.9:+COMP-NULL:+SALSA20-256:+UMAC-96:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION";
+				        req->selected_ciphersuite = "X-SALSA20-UMAC96";
+				        req->gnutls_cipher = GNUTLS_CIPHER_SALSA20_256;
+				        req->gnutls_mac = GNUTLS_MAC_UMAC_96;
+				        break;
+	                        } else
+#endif
+				if (strcmp(token, "AES128-SHA") == 0) {
+				        req->gnutls_ciphersuite = "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-128-CBC:+SHA1:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION";
+				        req->selected_ciphersuite = "AES128-SHA";
+				        req->gnutls_cipher = GNUTLS_CIPHER_AES_128_CBC;
+				        req->gnutls_mac = GNUTLS_MAC_SHA1;
+				        break;
+				} else if (strcmp(token, "DES-CBC3-SHA") == 0) {
+				        req->gnutls_ciphersuite = "NONE:+VERS-DTLS0.9:+COMP-NULL:+3DES-CBC:+SHA1:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION";
+				        req->selected_ciphersuite = "DES-CBC3-SHA";
+				        req->gnutls_cipher = GNUTLS_CIPHER_3DES_CBC;
+				        req->gnutls_mac = GNUTLS_MAC_SHA1;
+				        break;
+	                        }
+				str = NULL;
+			}
+
+			break;
+
+		case HEADER_CSTP_MTU:
+			req->cstp_mtu = atoi((char*)req->value.data);
+			break;
+		case HEADER_CSTP_ATYPE:
+			if (memmem(req->value.data, req->value.length, "IPv4", 4) == NULL)
+				req->no_ipv4 = 1;
+			if (memmem(req->value.data, req->value.length, "IPv6", 4) == NULL)
+				req->no_ipv6 = 1;
+			break;
+		case HEADER_DTLS_MTU:
+			req->dtls_mtu = atoi((char*)req->value.data);
+			break;
+		case HEADER_COOKIE:
+			length = req->value.length;
+			p = memmem(req->value.data, length, "webvpn=", 7);
+			if (p == NULL || length <= 7) {
+				req->cookie_set = 0;
+				return;
+			}
+			p += 7;
+			length -= 7;
+				
+			if (length < COOKIE_SIZE*2) {
+				req->cookie_set = 0;
+				return;
+			}
+			length = COOKIE_SIZE*2;
+			nlen = sizeof(req->cookie);
+			gnutls_hex2bin((void*)p, length, req->cookie, &nlen);
+
+			if (nlen < COOKIE_SIZE) {
+				req->cookie_set = 0;
+				return;
+			}
+			req->cookie_set = 1;
+			break;
+	}
+}
+
 int header_field_cb(http_parser* parser, const char *at, size_t length)
 {
 	struct worker_st *ws = parser->data;
 	struct http_req_st *req = &ws->req;
+	int ret;
 
-	if (length == sizeof(STR_HDR_COOKIE)-1 && strncmp(at, STR_HDR_COOKIE, length) == 0) {
+	if (req->header_state != HTTP_HEADER_RECV) {
+		/* handle value */
+		if (req->header_state == HTTP_HEADER_VALUE_RECV)
+			value_check(ws, req);
+		req->header_state = HTTP_HEADER_RECV;
+		str_reset(&req->header);
+	}
+
+	ret = str_append_data(&req->header, at, length);
+	if (ret < 0)
+		return ret;
+
+	return 0;
+}
+
+static void header_check(struct http_req_st *req)
+{
+	if (req->header.length == sizeof(STR_HDR_COOKIE)-1 && 
+	        strncmp((char*)req->header.data, STR_HDR_COOKIE, req->header.length) == 0) {
 		req->next_header = HEADER_COOKIE;
-	} else if (length == sizeof(STR_HDR_MS)-1 && strncmp(at, STR_HDR_MS, length) == 0) {
+	} else if (req->header.length == sizeof(STR_HDR_MS)-1 && 
+	        strncmp((char*)req->header.data, STR_HDR_MS, req->header.length) == 0) {
 		req->next_header = HEADER_MASTER_SECRET;
-	} else if (length == sizeof(STR_HDR_DMTU)-1 && strncmp(at, STR_HDR_DMTU, length) == 0) {
+	} else if (req->header.length == sizeof(STR_HDR_DMTU)-1 && 
+	        strncmp((char*)req->header.data, STR_HDR_DMTU, req->header.length) == 0) {
 		req->next_header = HEADER_DTLS_MTU;
-	} else if (length == sizeof(STR_HDR_CMTU)-1 && strncmp(at, STR_HDR_CMTU, length) == 0) {
+	} else if (req->header.length == sizeof(STR_HDR_CMTU)-1 && 
+	        strncmp((char*)req->header.data, STR_HDR_CMTU, req->header.length) == 0) {
 		req->next_header = HEADER_CSTP_MTU;
-	} else if (length == sizeof(STR_HDR_HOST)-1 && strncmp(at, STR_HDR_HOST, length) == 0) {
+	} else if (req->header.length == sizeof(STR_HDR_HOST)-1 && 
+	        strncmp((char*)req->header.data, STR_HDR_HOST, req->header.length) == 0) {
 		req->next_header = HEADER_HOSTNAME;
+	} else if (req->header.length == sizeof(STR_HDR_CS)-1 && 
+	        strncmp((char*)req->header.data, STR_HDR_CS, req->header.length) == 0) {
+		req->next_header = HEADER_DTLS_CIPHERSUITE;
+	} else if (req->header.length == sizeof(STR_HDR_ATYPE)-1 && 
+	        strncmp((char*)req->header.data, STR_HDR_ATYPE, req->header.length) == 0) {
+		req->next_header = HEADER_CSTP_ATYPE;
 	} else {
 		req->next_header = 0;
 	}
-	
-	if (length < sizeof(req->dbg_txt)) {
-		memcpy(req->dbg_txt, at, length);
-		req->dbg_txt[length] = 0;
-	} else {
-		oclog(ws, LOG_ERR, "oversized HTTP header %.*s\n", (int)length, at);
-		req->dbg_txt[0] = 0;
-	}
-	
-	return 0;
 }
 
 int header_value_cb(http_parser* parser, const char *at, size_t length)
 {
 	struct worker_st *ws = parser->data;
 	struct http_req_st *req = &ws->req;
-	char *p;
-	size_t nlen;
+	int ret;
 
-	if (length > 0) {
-		oclog(ws, LOG_DEBUG, "HTTP: %s: %.*s", req->dbg_txt, (int)length, at);
-
-		switch (req->next_header) {
-			case HEADER_MASTER_SECRET:
-				if (length < TLS_MASTER_SIZE*2) {
-					req->master_secret_set = 0;
-					return 0;
-				}
-				
-				length = TLS_MASTER_SIZE*2;
-
-				nlen = sizeof(req->master_secret);
-
-				gnutls_hex2bin(at, length, req->master_secret, &nlen);
-				req->master_secret_set = 1;
-				break;
-			case HEADER_HOSTNAME:
-				if (length >- MAX_HOSTNAME_SIZE) {
-					req->hostname[0] = 0;
-					return 0;
-				}
-				memcpy(req->hostname, at, length);
-				req->hostname[length] = 0;
-
-				break;
-			case HEADER_CSTP_MTU:
-				req->cstp_mtu = atoi(at);
-				break;
-			case HEADER_DTLS_MTU:
-				req->dtls_mtu = atoi(at);
-				break;
-			case HEADER_COOKIE:
-				p = memmem(at, length, "webvpn=", 7);
-				if (p == NULL || length <= 7) {
-					req->cookie_set = 0;
-					return 0;
-				}
-				p += 7;
-				length -= 7;
-				
-				if (length < COOKIE_SIZE*2) {
-					req->cookie_set = 0;
-					return 0;
-				}
-				length = COOKIE_SIZE*2;
-
-				nlen = sizeof(req->cookie);
-				gnutls_hex2bin(p, length, req->cookie, &nlen);
-
-				if (nlen < COOKIE_SIZE) {
-					req->cookie_set = 0;
-					return 0;
-				}
-				req->cookie_set = 1;
-				break;
-		}
+	if (req->header_state != HTTP_HEADER_VALUE_RECV) {
+		/* handle header */
+		header_check(req);
+		req->header_state = HTTP_HEADER_VALUE_RECV;
+		str_reset(&req->value);
 	}
-	
+
+	ret = str_append_data(&req->value, at, length);
+	if (ret < 0)
+		return ret;
+
 	return 0;
 }
 
@@ -354,6 +361,9 @@ int header_complete_cb(http_parser* parser)
 	struct worker_st *ws = parser->data;
 	struct http_req_st *req = &ws->req;
 
+	/* handle header value */
+	value_check(ws, req);
+
 	req->headers_complete = 1;
 	return 0;
 }
@@ -385,9 +395,6 @@ int body_cb(http_parser* parser, const char *at, size_t length)
 	return 0;
 }
 
-#define GNUTLS_CIPHERSUITE "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-128-CBC:+SHA1:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION"
-#define OPENSSL_CIPHERSUITE "AES128-SHA"
-
 static int setup_dtls_connection(struct worker_st *ws)
 {
 int ret;
@@ -395,6 +402,12 @@ gnutls_session_t session;
 gnutls_datum_t master = { ws->master_secret, sizeof(ws->master_secret) };
 gnutls_datum_t sid = { ws->session_id, sizeof(ws->session_id) };
 
+        if (ws->req.gnutls_ciphersuite == NULL) {
+		oclog(ws, LOG_ERR, "no DTLS ciphersuite negotiated");
+		return -1;
+        }
+
+	oclog(ws, LOG_INFO, "setting up DTLS connection");
 	/* DTLS cookie verified.
 	 * Initialize session.
 	 */
@@ -404,15 +417,15 @@ gnutls_datum_t sid = { ws->session_id, sizeof(ws->session_id) };
 		return -1;
 	}
 
-	ret = gnutls_priority_set_direct(session, GNUTLS_CIPHERSUITE, NULL);
+	ret = gnutls_priority_set_direct(session, ws->req.gnutls_ciphersuite, NULL);
 	if (ret < 0) {
 		oclog(ws, LOG_ERR, "could not set TLS priority: %s", gnutls_strerror(ret));
 		goto fail;
 	}
 
 	ret = gnutls_session_set_premaster(session, GNUTLS_SERVER,
-		GNUTLS_DTLS0_9, GNUTLS_KX_RSA, GNUTLS_CIPHER_AES_128_CBC,
-		GNUTLS_MAC_SHA1, GNUTLS_COMP_NULL, &master, &sid);
+		GNUTLS_DTLS0_9, GNUTLS_KX_RSA, ws->req.gnutls_cipher,
+		ws->req.gnutls_mac, GNUTLS_COMP_NULL, &master, &sid);
 	if (ret < 0) {
 		oclog(ws, LOG_ERR, "could not set TLS premaster: %s", gnutls_strerror(ret));
 		goto fail;
@@ -442,18 +455,29 @@ fail:
 	return -1;
 }
 
+static void http_req_init(worker_st * ws)
+{
+	str_init(&ws->req.header);
+	str_init(&ws->req.value);
+}
+
 static void http_req_reset(worker_st * ws)
 {
 	ws->req.headers_complete = 0;
 	ws->req.message_complete = 0;
 	ws->req.body_length = 0;
 	ws->req.url[0] = 0;
-	ws->req.dbg_txt[0] = 0;
+	
+	ws->req.header_state = HTTP_HEADER_INIT;
+	str_reset(&ws->req.header);
+	str_reset(&ws->req.value);
 }
 
 static void http_req_deinit(worker_st * ws)
 {
 	http_req_reset(ws);
+	str_clear(&ws->req.header);
+	str_clear(&ws->req.value);
 	free(ws->req.body);
 	ws->req.body = NULL;
 }
@@ -476,18 +500,17 @@ void vpn_server(struct worker_st* ws)
 	url_handler_fn fn;
 	int requests_left = MAX_HTTP_REQUESTS;
 
-	signal(SIGTERM, handle_term);
-	signal(SIGINT, handle_term);
-	signal(SIGHUP, SIG_IGN);
-	signal(SIGALRM, handle_alarm);
+	ocsignal(SIGTERM, handle_term);
+	ocsignal(SIGINT, handle_term);
+	ocsignal(SIGHUP, SIG_IGN);
+	ocsignal(SIGALRM, handle_alarm);
 
 	if (ws->config->auth_timeout)
 		alarm(ws->config->auth_timeout);
 		
 	ret = disable_system_calls(ws);
 	if (ret < 0) {
-		oclog(ws, LOG_ERR, "could not disable system calls (seccomp error)");
-		exit_worker(ws);
+		oclog(ws, LOG_INFO, "could not disable system calls, kernel might not support seccomp");
 	}
 
 	oclog(ws, LOG_INFO, "accepted connection");
@@ -531,6 +554,7 @@ void vpn_server(struct worker_st* ws)
 	settings.on_headers_complete = header_complete_cb;
 	settings.on_message_complete = message_complete_cb;
 	settings.on_body = body_cb;
+	http_req_init(ws);
 
 	ws->session = session;
 	ws->parser = &parser;
@@ -544,16 +568,16 @@ restart:
 	http_parser_init(&parser, HTTP_REQUEST);
 	parser.data = ws;
 	http_req_reset(ws);
-
-
 	/* parse as we go */
 	do {
 		nrecvd = tls_recv(session, buf, sizeof(buf));
 		if (nrecvd <= 0) {
+			if (nrecvd == 0)
+			        goto finish;
 			oclog(ws, LOG_INFO, "error receiving client data"); 
 			exit_worker(ws);
 		}
-	
+
 		nparsed = http_parser_execute(&parser, &settings, (void*)buf, nrecvd);
 		if (nparsed == 0) {
 			oclog(ws, LOG_INFO, "error parsing HTTP request"); 
@@ -565,11 +589,10 @@ restart:
 		oclog(ws, LOG_DEBUG, "HTTP GET %s", ws->req.url); 
 		fn = get_url_handler(ws->req.url);
 		if (fn == NULL) {
-			oclog(ws, LOG_INFO, "unexpected URL %s", ws->req.url); 
-			tls_puts(session, "HTTP/1.1 404 Nah, go away\r\n\r\n");
+			oclog(ws, LOG_INFO, "unexpected URL %s", ws->req.url);
+			tls_puts(session, "HTTP/1.1 404 Not found\r\n\r\n");
 			goto finish;
-		}
-		
+                }		
 		ret = fn(ws, parser.http_minor);
 		if (ret == 0 && (parser.http_major != 1 || parser.http_minor != 0))
 			goto restart;
@@ -591,7 +614,7 @@ restart:
 		fn = post_url_handler(ws->req.url);
 		if (fn == NULL) {
 			oclog(ws, LOG_INFO, "unexpected POST URL %s", ws->req.url); 
-			tls_printf(session, "HTTP/1.%u 404 Nah, go away\r\n\r\n", parser.http_minor);
+			tls_puts(session, "HTTP/1.1 404 Not found\r\n\r\n");
 			goto finish;
 		}
 
@@ -614,7 +637,18 @@ finish:
 	tls_close(session);
 }
 
-/* sets the provided value of mtu as bad,
+static
+void mtu_set(worker_st* ws, unsigned mtu)
+{
+	ws->conn_mtu = mtu;
+	
+	if (ws->dtls_session)
+		gnutls_dtls_set_data_mtu (ws->dtls_session, mtu);
+	send_tun_mtu(ws, mtu - 1); /* for DTLS header */
+	oclog(ws, LOG_INFO, "setting MTU to %u", ws->conn_mtu);
+}
+
+/* sets the current value of mtu as bad,
  * and returns an estimation of good.
  *
  * Returns -1 on failure.
@@ -622,10 +656,10 @@ finish:
 static
 int mtu_not_ok(worker_st* ws)
 {
-	ws->last_bad_mtu = ws->dtls_mtu;
+	ws->last_bad_mtu = ws->conn_mtu;
 
-	if (ws->last_good_mtu >= ws->dtls_mtu) {
-		ws->last_good_mtu = (ws->dtls_mtu)/2;
+	if (ws->last_good_mtu >= ws->conn_mtu) {
+		ws->last_good_mtu = (2*(ws->conn_mtu))/3;
 	
 		if (ws->last_good_mtu < 128) {
 			oclog(ws, LOG_INFO, "could not calculate a valid MTU. Disabling DTLS.");
@@ -634,20 +668,16 @@ int mtu_not_ok(worker_st* ws)
 		}
 	}
 
-	ws->dtls_mtu = ws->last_good_mtu;
-	gnutls_dtls_set_data_mtu (ws->dtls_session, ws->dtls_mtu);
-
-	oclog(ws, LOG_DEBUG, "MTU %u is too large, switching to %u", ws->last_bad_mtu, ws->dtls_mtu);
-
-	send_tun_mtu(ws, ws->dtls_mtu-1);
+	mtu_set(ws, ws->last_good_mtu);
+	oclog(ws, LOG_INFO, "MTU %u is too large, switching to %u", ws->last_bad_mtu, ws->conn_mtu);
 
 	return 0;
 }
 
-static void mtu_discovery_init(worker_st *ws)
+static void mtu_discovery_init(worker_st *ws, unsigned mtu)
 {
-	ws->last_good_mtu = ws->dtls_mtu;
-	ws->last_bad_mtu = ws->dtls_mtu;
+	ws->last_good_mtu = mtu;
+	ws->last_bad_mtu = mtu;
 }
 
 static
@@ -655,36 +685,86 @@ void mtu_ok(worker_st* ws)
 {
 unsigned int c;
 
-	if (ws->last_bad_mtu == (ws->dtls_mtu)+1 ||
-		ws->last_bad_mtu == (ws->dtls_mtu))
+	if (ws->last_bad_mtu == (ws->conn_mtu)+1 ||
+		ws->last_bad_mtu == (ws->conn_mtu))
 		return;
 	
-	ws->last_good_mtu = ws->dtls_mtu;
-	c = (ws->dtls_mtu + ws->last_bad_mtu)/2;
-
-	ws->dtls_mtu = c;
-	gnutls_dtls_set_data_mtu (ws->dtls_session, c);
-	send_tun_mtu(ws, c-1);
-
-	oclog(ws, LOG_DEBUG, "trying MTU %u", c);
+	ws->last_good_mtu = ws->conn_mtu;
+	c = (ws->conn_mtu + ws->last_bad_mtu)/2;
 
+	mtu_set(ws, c);
 	return;
 }
 
+static 
+int periodic_check(worker_st *ws, unsigned mtu_overhead, time_t now)
+{
+socklen_t sl;
+int max, e, ret;
+
+	if (now - ws->last_periodic_check < PERIODIC_CHECK_TIME)
+		return 0;
+
+	/* check DPD. Otherwise exit */
+	if (ws->udp_state == UP_ACTIVE && now-ws->last_msg_udp > DPD_TRIES*ws->config->dpd) {
+		oclog(ws, LOG_ERR, "have not received UDP any message or DPD for long (%d secs)", (int)(now-ws->last_msg_udp));
+				
+		ws->buffer[0] = AC_PKT_DPD_OUT;
+		tls_send(ws->dtls_session, ws->buffer, 1);
+				
+		if (now-ws->last_msg_udp > DPD_MAX_TRIES*ws->config->dpd) {
+			oclog(ws, LOG_ERR, "have not received UDP message or DPD for very long; disabling UDP port");
+			ws->udp_state = UP_INACTIVE;
+		}
+	}
+	if (now-ws->last_msg_tcp > DPD_TRIES*ws->config->dpd) {
+		oclog(ws, LOG_ERR, "have not received TCP DPD for long (%d secs)", (int)(now-ws->last_msg_tcp));
+		ws->buffer[0] = AC_PKT_DPD_OUT;
+		tls_send(ws->session, ws->buffer, 1);
+
+		if (now-ws->last_msg_tcp > DPD_MAX_TRIES*ws->config->dpd) {
+			oclog(ws, LOG_ERR, "have not received TCP DPD for very long; tearing down connection");
+			return -1;
+		}
+	}
+
+	sl = sizeof(max);
+	ret = getsockopt(ws->conn_fd, IPPROTO_TCP, TCP_MAXSEG, &max, &sl);
+	if (ret == -1) {
+		e = errno;
+		oclog(ws, LOG_INFO, "error in getting TCP_MAXSEG: %s", strerror(e));
+	} else {
+		max -= 13;
+		oclog(ws, LOG_DEBUG, "TCP MSS is %u", max);
+		if (max > 0 && max-mtu_overhead < ws->conn_mtu) {
+			oclog(ws, LOG_INFO, "reducing MTU due to TCP MSS to %u", max-mtu_overhead);
+			mtu_set(ws, MIN(ws->conn_mtu, max-mtu_overhead));
+		}
+	}
+
+
+	ws->last_periodic_check = now;
+	
+	return 0;
+}
+
+#define CSTP_DTLS_OVERHEAD 1
+#define CSTP_OVERHEAD 8
 
 #define SEND_ERR(x) if (x<0) goto send_error
 static int connect_handler(worker_st *ws)
 {
 struct http_req_st *req = &ws->req;
 fd_set rfds;
-int l, e, max, ret;
+int l, e, max, ret, overhead;
 struct vpn_st vinfo;
-unsigned tls_retry;
+unsigned tls_retry, dtls_mtu, cstp_mtu;
 char *p;
 struct timeval tv;
 unsigned tls_pending, dtls_pending = 0, i;
 time_t udp_recv_time = 0, now;
-unsigned mtu_overhead, tls_mtu = 0;
+unsigned mtu_overhead = 0;
+socklen_t sl;
 
 	ws->buffer_size = 16*1024;
 	ws->buffer = malloc(ws->buffer_size);
@@ -695,14 +775,14 @@ unsigned mtu_overhead, tls_mtu = 0;
 		exit_worker(ws);
 	}
 
-	if (req->cookie_set == 0) {
+        if (ws->auth_state != S_AUTH_COMPLETE && req->cookie_set == 0) {
 		oclog(ws, LOG_INFO, "connect request without authentication");
 		tls_puts(ws->session, "HTTP/1.1 503 Service Unavailable\r\n\r\n");
 		tls_fatal_close(ws->session, GNUTLS_A_ACCESS_DENIED);
 		exit_worker(ws);
 	}
 
-	if (ws->auth_ok == 0) {
+	if (ws->auth_state != S_AUTH_COMPLETE) {
 		/* authentication didn't occur in this session. Use the
 		 * cookie */
 		ret = auth_cookie(ws, req->cookie, sizeof(req->cookie));
@@ -740,7 +820,7 @@ unsigned mtu_overhead, tls_mtu = 0;
 	if (ws->config->auth_timeout)
 		alarm(0);
 	http_req_deinit(ws);
-	
+
 	tls_cork(ws->session);
 	ret = tls_puts(ws->session, "HTTP/1.1 200 CONNECTED\r\n");
 	SEND_ERR(ret);
@@ -751,6 +831,11 @@ unsigned mtu_overhead, tls_mtu = 0;
 	ret = tls_printf(ws->session, "X-CSTP-DPD: %u\r\n", ws->config->dpd);
 	SEND_ERR(ret);
 
+	if (ws->config->default_domain) {
+        	ret = tls_printf(ws->session, "X-CSTP-Default-Domain: %s\r\n", ws->config->default_domain);
+        	SEND_ERR(ret);
+        }
+
 	ws->udp_state = UP_DISABLED;
 	if (req->master_secret_set != 0) {
 		memcpy(ws->master_secret, req->master_secret, TLS_MASTER_SIZE);
@@ -759,7 +844,7 @@ unsigned mtu_overhead, tls_mtu = 0;
 		oclog(ws, LOG_DEBUG, "disabling UDP (DTLS) connection");
 	}
 
-	if (vinfo.ipv4) {
+	if (vinfo.ipv4 && req->no_ipv4 == 0) {
 		oclog(ws, LOG_DEBUG, "sending IPv4 %s", vinfo.ipv4);
 		ret = tls_printf(ws->session, "X-CSTP-Address: %s\r\n", vinfo.ipv4);
 		SEND_ERR(ret);
@@ -778,7 +863,7 @@ unsigned mtu_overhead, tls_mtu = 0;
 		}
 	}
 	
-	if (vinfo.ipv6) {
+	if (vinfo.ipv6 && req->no_ipv6 == 0) {
 		oclog(ws, LOG_DEBUG, "sending IPv6 %s", vinfo.ipv6);
 		ret = tls_printf(ws->session, "X-CSTP-Address: %s\r\n", vinfo.ipv6);
 		SEND_ERR(ret);
@@ -798,6 +883,10 @@ unsigned mtu_overhead, tls_mtu = 0;
 	}
 
 	for (i=0;i<vinfo.routes_size;i++) {
+		if (req->no_ipv6 != 0 && strchr(vinfo.routes[i], ':') != 0)
+			continue;
+		if (req->no_ipv4 != 0 && strchr(vinfo.routes[i], '.') != 0)
+			continue;
 		oclog(ws, LOG_DEBUG, "adding route %s", vinfo.routes[i]);
 		ret = tls_printf(ws->session,
 			"X-CSTP-Split-Include: %s\r\n", vinfo.routes[i]);
@@ -823,18 +912,34 @@ unsigned mtu_overhead, tls_mtu = 0;
 		);
 	SEND_ERR(ret);
 
+	if (ws->config->default_mtu > 0) {
+		vinfo.mtu = ws->config->default_mtu;
+	}
 
-	tls_mtu = vinfo.mtu - 8;
+	mtu_overhead = CSTP_OVERHEAD;
+	ws->conn_mtu = vinfo.mtu - mtu_overhead;
 	if (req->cstp_mtu > 0) {
-		tls_mtu = MIN(tls_mtu, req->cstp_mtu);
+		ws->conn_mtu = MIN(ws->conn_mtu, req->cstp_mtu);
 		oclog(ws, LOG_DEBUG, "peer CSTP MTU is %u", req->cstp_mtu);
 	}
-	tls_mtu = MIN(ws->buffer_size-8, tls_mtu);
 
-	ret = tls_printf(ws->session, "X-CSTP-MTU: %u\r\n", tls_mtu);
-	SEND_ERR(ret);
+	sl = sizeof(max);
+	ret = getsockopt(ws->conn_fd, IPPROTO_TCP, TCP_MAXSEG, &max, &sl);
+	if (ret == -1) {
+		e = errno;
+		oclog(ws, LOG_INFO, "error in getting TCP_MAXSEG: %s", strerror(e));
+	} else {
+		max -= 13;
+		oclog(ws, LOG_INFO, "TCP MSS is %u", max);
+		if (max > 0 && max-mtu_overhead < ws->conn_mtu) {
+			oclog(ws, LOG_DEBUG, "reducing MTU due to TCP MSS to %u", max-mtu_overhead);
+		}
+		ws->conn_mtu = MIN(ws->conn_mtu, max-mtu_overhead);
+	}
 
 	if (ws->udp_state != UP_DISABLED) {
+		int sndbuf;
+
 		p = (char*)ws->buffer;
 		for (i=0;i<sizeof(ws->session_id);i++) {
 			sprintf(p, "%.2x", (unsigned int)ws->session_id[i]);
@@ -855,30 +960,55 @@ unsigned mtu_overhead, tls_mtu = 0;
 		ret = tls_printf(ws->session, "X-DTLS-Keepalive: %u\r\n", ws->config->keepalive);
 		SEND_ERR(ret);
 
-		ret = tls_puts(ws->session, "X-DTLS-CipherSuite: "OPENSSL_CIPHERSUITE"\r\n");
+		oclog(ws, LOG_INFO, "DTLS ciphersuite: %s", ws->req.selected_ciphersuite);
+		ret = tls_printf(ws->session, "X-DTLS-CipherSuite: %s\r\n", ws->req.selected_ciphersuite);
 		SEND_ERR(ret);
 
 		/* assume that if IPv6 is used over TCP then the same would be used over UDP */
 		if (ws->proto == AF_INET)
-			mtu_overhead = 20+1; /* ip */
+			mtu_overhead = 20+CSTP_DTLS_OVERHEAD; /* ip */
 		else
-			mtu_overhead = 40+1; /* ipv6 */
+			mtu_overhead = 40+CSTP_DTLS_OVERHEAD; /* ipv6 */
 		mtu_overhead += 8; /* udp */
-		ws->dtls_mtu = vinfo.mtu - mtu_overhead;
+		ws->conn_mtu = MIN(ws->conn_mtu, vinfo.mtu - mtu_overhead);
 
 		if (req->dtls_mtu > 0) {
-			ws->dtls_mtu = MIN(req->dtls_mtu, ws->dtls_mtu);
-			oclog(ws, LOG_DEBUG, "peer DTLS MTU is %u", req->dtls_mtu);
+			ws->conn_mtu = MIN(req->dtls_mtu, ws->conn_mtu);
+			oclog(ws, LOG_INFO, "reducing DTLS MTU to peer's DTLS MTU (%u)", req->dtls_mtu);
 		}
 
-		ws->dtls_mtu = MIN(ws->buffer_size-1, ws->dtls_mtu);
-		tls_printf(ws->session, "X-DTLS-MTU: %u\r\n", ws->dtls_mtu);
+		overhead = CSTP_DTLS_OVERHEAD + tls_get_overhead(GNUTLS_DTLS0_9, ws->req.gnutls_cipher, ws->req.gnutls_mac);
+		dtls_mtu = ws->conn_mtu - overhead;
+
+		tls_printf(ws->session, "X-DTLS-MTU: %u\r\n", dtls_mtu);
+		oclog(ws, LOG_INFO, "suggesting DTLS MTU %u", dtls_mtu);
+		
+		if (ws->config->output_buffer > 0) {
+			sndbuf = ws->conn_mtu * ws->config->output_buffer;
+			setsockopt( ws->udp_fd, SOL_SOCKET, SO_SNDBUF, &sndbuf, sizeof(sndbuf));
+		}
+	} else
+		dtls_mtu = 0;
+	
+	if (ws->buffer_size <= ws->conn_mtu+mtu_overhead) {
+		oclog(ws, LOG_WARNING, "buffer size is smaller than MTU (%u < %u); adjusting", ws->buffer_size, ws->conn_mtu);
+		ws->buffer_size = ws->conn_mtu+mtu_overhead;
+		ws->buffer = realloc(ws->buffer, ws->buffer_size);
+		if (ws->buffer == NULL)
+			goto exit;
 	}
 
-	if (ws->dtls_mtu == 0)
-		send_tun_mtu(ws, tls_mtu);
-	else
-		send_tun_mtu(ws, ws->dtls_mtu-1);
+	overhead = CSTP_OVERHEAD + tls_get_overhead(gnutls_protocol_get_version(ws->session), gnutls_cipher_get(ws->session), gnutls_mac_get(ws->session));
+	cstp_mtu = ws->conn_mtu-overhead;
+	if (dtls_mtu > 0) /* this is a hack for openconnect which reads a single MTU value */
+		cstp_mtu = MIN(cstp_mtu, dtls_mtu);
+
+	ret = tls_printf(ws->session, "X-CSTP-MTU: %u\r\n", cstp_mtu);
+	SEND_ERR(ret);
+	oclog(ws, LOG_INFO, "suggesting CSTP MTU %u", cstp_mtu);
+
+	oclog(ws, LOG_INFO, "plaintext MTU is %u", ws->conn_mtu);
+	send_tun_mtu(ws, ws->conn_mtu);
 
 	if (ws->config->banner) {
 		ret = tls_printf(ws->session, "X-CSTP-Banner: %s\r\n", ws->config->banner);
@@ -892,7 +1022,7 @@ unsigned mtu_overhead, tls_mtu = 0;
 	SEND_ERR(ret);
 
 	/* start dead peer detection */
-	ws->last_dpd_tcp = ws->last_dpd_udp = time(0);
+	ws->last_msg_tcp = ws->last_msg_udp = time(0);
 
 	/* main loop  */
 	for(;;) {
@@ -910,15 +1040,6 @@ unsigned mtu_overhead, tls_mtu = 0;
 		}
 
 		if (terminate != 0) {
-			if (ws->udp_state == UP_ACTIVE) {
-				ws->buffer[0] = AC_PKT_TERM_SERVER;
-				
-				oclog(ws, LOG_DEBUG, "sending disconnect message in DTLS channel");
-
-				ret = tls_send(ws->dtls_session, ws->buffer, 1);
-				GNUTLS_FATAL_ERR(ret);
-			}
-
 			ws->buffer[0] = 'S';
 			ws->buffer[1] = 'T';
 			ws->buffer[2] = 'F';
@@ -931,9 +1052,9 @@ unsigned mtu_overhead, tls_mtu = 0;
 			oclog(ws, LOG_DEBUG, "sending disconnect message in TLS channel");
 			ret = tls_send(ws->session, ws->buffer, 8);
 			GNUTLS_FATAL_ERR(ret);
-
 			goto exit;
 		}
+		
 
 		tls_pending = gnutls_record_check_pending(ws->session);
 
@@ -948,43 +1069,13 @@ unsigned mtu_overhead, tls_mtu = 0;
 					continue;
 				goto exit;
 			}
-			
-			if (ret == 0) {
-				now = time(0);
-				/* check DPD. Otherwise exit */
-				if (ws->udp_state == UP_ACTIVE && now-ws->last_dpd_udp > DPD_TRIES*ws->config->dpd) {
-					oclog(ws, LOG_ERR, "have not received UDP DPD for long (%d secs)", (int)(now-ws->last_dpd_udp));
-				
-					ws->buffer[0] = AC_PKT_DPD_OUT;
-					tls_send(ws->dtls_session, ws->buffer, 1);
-				
-					if (now-ws->last_dpd_udp > DPD_MAX_TRIES*ws->config->dpd) {
-						oclog(ws, LOG_ERR, "have not received UDP DPD for very long; disabling UDP port");
-						ws->udp_state = UP_INACTIVE;
-					}
-				}
-				if (now-ws->last_dpd_tcp > DPD_TRIES*ws->config->dpd) {
-					oclog(ws, LOG_ERR, "have not received TCP DPD for long (%d secs)", (int)(now-ws->last_dpd_tcp));
-					ws->buffer[0] = AC_PKT_DPD_OUT;
-					tls_send(ws->session, ws->buffer, 1);
-
-					if (now-ws->last_dpd_tcp > DPD_MAX_TRIES*ws->config->dpd) {
-						oclog(ws, LOG_ERR, "have not received TCP DPD for very long; tearing down connection");
-						goto exit;
-					}
-				}
-			}
 		}
+		now = time(0);
+		if (periodic_check(ws, mtu_overhead, now) < 0)
+			goto exit;
 		
 		if (FD_ISSET(ws->tun_fd, &rfds)) {
-				
-			if (ws->udp_state == UP_ACTIVE) {
-				l = ws->dtls_mtu-1;
-			} else {
-				l = tls_mtu;
-			}
-
-			l = read(ws->tun_fd, ws->buffer + 8, l);
+			l = read(ws->tun_fd, ws->buffer + 8, ws->conn_mtu - 1);
 			if (l < 0) {
 				e = errno;
 				
@@ -995,9 +1086,9 @@ unsigned mtu_overhead, tls_mtu = 0;
 				continue;
 			}
 			
-			if (l == 0) { /* disconnect */
+			if (l == 0) {
 				oclog(ws, LOG_INFO, "TUN device returned zero");
-				goto exit;
+				continue;
 			}
 
 			tls_retry = 0;
@@ -1005,7 +1096,7 @@ unsigned mtu_overhead, tls_mtu = 0;
 			if (ws->udp_state == UP_ACTIVE) {
 				ws->buffer[7] = AC_PKT_DATA;
 
-				ret = tls_send(ws->dtls_session, ws->buffer + 7, l+1);
+				ret = tls_send_nowait(ws->dtls_session, ws->buffer + 7, l + 1);
 				GNUTLS_FATAL_ERR(ret);
 
 				if (ret == GNUTLS_E_LARGE_PACKET) {
@@ -1013,6 +1104,8 @@ unsigned mtu_overhead, tls_mtu = 0;
 
 					oclog(ws, LOG_DEBUG, "retrying (TLS) %d\n", l);
 					tls_retry = 1;
+				} else if (ret >= ws->conn_mtu && ws->config->try_mtu != 0) {
+					mtu_ok(ws);
 				}
 			}
 
@@ -1046,7 +1139,7 @@ unsigned mtu_overhead, tls_mtu = 0;
 			if (ret > 0) {
 				l = ret;
 
-				ret = parse_cstp_data(ws, ws->buffer, l);
+				ret = parse_cstp_data(ws, ws->buffer, l, now);
 				if (ret < 0) {
 					oclog(ws, LOG_INFO, "error parsing CSTP data");
 					goto exit;
@@ -1074,7 +1167,7 @@ unsigned mtu_overhead, tls_mtu = 0;
 						l = ret;
 						ws->udp_state = UP_ACTIVE;
 
-						ret = parse_dtls_data(ws, ws->buffer, l);
+						ret = parse_dtls_data(ws, ws->buffer, l, now);
 						if (ret < 0) {
 							oclog(ws, LOG_INFO, "error parsing CSTP data");
 							goto exit;
@@ -1090,8 +1183,8 @@ unsigned mtu_overhead, tls_mtu = 0;
 					if (ret < 0)
 						goto exit;
 					
-					gnutls_dtls_set_mtu (ws->dtls_session, ws->dtls_mtu);
-					mtu_discovery_init(ws);
+					gnutls_dtls_set_mtu (ws->dtls_session, ws->conn_mtu);
+					mtu_discovery_init(ws, ws->conn_mtu);
 
 					break;
 				case UP_HANDSHAKE:
@@ -1112,14 +1205,14 @@ hsk_restart:
 						if (ret == 0) {
 							goto hsk_restart;
 						}
-
 					}
 
 					if (ret == 0) {
+						unsigned mtu = gnutls_dtls_get_data_mtu(ws->dtls_session);
 						ws->udp_state = UP_ACTIVE;
-						ws->dtls_mtu = gnutls_dtls_get_data_mtu(ws->dtls_session);
-						mtu_discovery_init(ws);
-						oclog(ws, LOG_DEBUG, "DTLS handshake completed (MTU: %u)\n", ws->dtls_mtu);
+						mtu_discovery_init(ws, mtu);
+						mtu_set(ws, mtu);
+						oclog(ws, LOG_INFO, "DTLS handshake completed (MTU: %u)\n", ws->conn_mtu);
 					}
 					
 					break;
@@ -1163,7 +1256,7 @@ static int parse_data(struct worker_st* ws,
 			uint8_t head,
 			uint8_t* buf, size_t buf_size, time_t now)
 {
-int ret, e, l;
+int ret, e;
 
 	switch (head) {
 		case AC_PKT_DPD_RESP:
@@ -1181,24 +1274,10 @@ int ret, e, l;
 				/* Use DPD for MTU discovery in DTLS */
 				ws->buffer[0] = AC_PKT_DPD_RESP;
 				
-				if (ws->config->try_mtu == 0 || ws->dpd_mtu_trial == 0) {
-					l = 1;
-					if (now-ws->last_dpd_udp <= ws->config->dpd/2)
-						mtu_not_ok(ws);
-
-					ws->dpd_mtu_trial++;
-				} else {
-					l = ws->dtls_mtu;
-					ws->dpd_mtu_trial = 0;
-				}
-				ret = tls_send(ts, ws->buffer, l);
-
+				ret = tls_send(ts, ws->buffer, 1);
 				if (ret == GNUTLS_E_LARGE_PACKET) {
 					mtu_not_ok(ws);
-					tls_send(ts, ws->buffer, 1);
-					ret = 1;
-				} else if (ws->config->try_mtu != 0 && ret > 0 && ws->dpd_mtu_trial > 0) {
-					mtu_ok(ws);
+					ret = tls_send(ts, ws->buffer, 1);
 				}
 
 				oclog(ws, LOG_DEBUG, "received DTLS DPD; sent response (%d bytes)", ret);
@@ -1215,7 +1294,7 @@ int ret, e, l;
 			break;
 		case AC_PKT_DATA:
 			oclog(ws, LOG_DEBUG, "writing %d byte(s) to TUN", (int)buf_size);
-			ret = tun_write(ws->tun_fd, buf, buf_size);
+			ret = force_write(ws->tun_fd, buf, buf_size);
 			if (ret == -1) {
 				e = errno;
 				oclog(ws, LOG_ERR, "could not write data to tun: %s", strerror(e));
@@ -1231,10 +1310,9 @@ int ret, e, l;
 }
 
 static int parse_cstp_data(struct worker_st* ws, 
-				uint8_t* buf, size_t buf_size)
+				uint8_t* buf, size_t buf_size, time_t now)
 {
-int pktlen;
-time_t now;
+int pktlen, ret;
 
 	if (buf_size < 8) {
 		oclog(ws, LOG_INFO, "can't read CSTP header (only %d bytes are available)\n", (int)buf_size);
@@ -1253,24 +1331,25 @@ time_t now;
 		return -1;
 	}
 
+	ret = parse_data(ws, ws->session, buf[6], buf+8, pktlen, now);
 	/* whatever we received treat it as DPD response.
 	 * it indicates that the channel is alive */
-	now = time(0);
-	ws->last_dpd_tcp = now;
-	return parse_data(ws, ws->session, buf[6], buf+8, pktlen, now);
+	ws->last_msg_tcp = now;
+	
+	return ret;
 }
 
 static int parse_dtls_data(struct worker_st* ws, 
-				uint8_t* buf, size_t buf_size)
+				uint8_t* buf, size_t buf_size, time_t now)
 {
-time_t now;
+int ret;
 
 	if (buf_size < 1) {
 		oclog(ws, LOG_INFO, "can't read DTLS header (only %d bytes are available)\n", (int)buf_size);
 		return -1;
 	}
 
-	now = time(0);
-	ws->last_dpd_udp = now;
-	return parse_data(ws, ws->dtls_session, buf[0], buf+1, buf_size-1, now);
+	ret = parse_data(ws, ws->dtls_session, buf[0], buf+1, buf_size-1, now);
+	ws->last_msg_udp = now;
+	return ret;
 }
