@@ -10,6 +10,7 @@
 #include <vpn.h>
 #include <cookies.h>
 #include <tlslib.h>
+#include <str.h>
 
 typedef enum {
 	UP_DISABLED,
@@ -25,12 +26,30 @@ enum {
 	HEADER_MASTER_SECRET,
 	HEADER_HOSTNAME,
 	HEADER_CSTP_MTU,
+	HEADER_CSTP_ATYPE,
 	HEADER_DTLS_MTU,
+	HEADER_DTLS_CIPHERSUITE,
+};
+
+enum {
+	HTTP_HEADER_INIT = 0,
+	HTTP_HEADER_RECV,
+	HTTP_HEADER_VALUE_RECV
+};
+
+enum {
+	S_AUTH_INACTIVE = 0,
+	S_AUTH_INIT,
+	S_AUTH_REQ,
+	S_AUTH_COMPLETE
 };
 
 struct http_req_st {
 	char url[256];
-	char dbg_txt[256];
+
+	str_st header;
+	str_st value;
+	unsigned int header_state;
 
 	char hostname[MAX_HOSTNAME_SIZE];
 	unsigned int next_header;
@@ -41,10 +60,19 @@ struct http_req_st {
 
 	char *body;
 	unsigned int body_length;
+
+	char *gnutls_ciphersuite; /* static string */
+	char *selected_ciphersuite; /* static string */
+	int gnutls_cipher;
+	int gnutls_mac;
+
 	unsigned int headers_complete;
 	unsigned int message_complete;
 	unsigned dtls_mtu;
 	unsigned cstp_mtu;
+	
+	unsigned no_ipv4;
+	unsigned no_ipv6;
 };
 
 typedef struct worker_st {
@@ -56,14 +84,16 @@ typedef struct worker_st {
 	
 	http_parser *parser;
 	struct cfg_st *config;
+	unsigned int auth_state; /* S_AUTH */
 
 	struct sockaddr_storage remote_addr;	/* peer's address */
 	socklen_t remote_addr_len;
 	int proto; /* AF_INET or AF_INET6 */
 	
 	/* for dead peer detection */
-	time_t last_dpd_udp;
-	time_t last_dpd_tcp;
+	time_t last_msg_udp;
+	time_t last_msg_tcp;
+	time_t last_periodic_check;
 
 	/* set after authentication */
 	int udp_fd;
@@ -72,8 +102,7 @@ typedef struct worker_st {
 	/* for mtu trials */
 	unsigned last_good_mtu;
 	unsigned last_bad_mtu;
-	unsigned dtls_mtu;
-	unsigned dpd_mtu_trial;
+	unsigned conn_mtu;
 	
 	/* Buffer used by worker */
 	uint8_t * buffer;
@@ -86,7 +115,6 @@ typedef struct worker_st {
 	uint8_t cookie[COOKIE_SIZE];
 	uint8_t master_secret[TLS_MASTER_SIZE];
 	uint8_t session_id[GNUTLS_MAX_SESSION_ID];
-	unsigned auth_ok;
 	unsigned cert_auth_ok;
 	int tun_fd;
 	
@@ -96,10 +124,16 @@ typedef struct worker_st {
 void vpn_server(struct worker_st* ws);
 
 int auth_cookie(worker_st *ws, void* cookie, size_t cookie_size);
+int auth_user_deinit(worker_st *ws);
 
 int get_auth_handler(worker_st *server, unsigned http_ver);
 int post_auth_handler(worker_st *server, unsigned http_ver);
 
+int get_empty_handler(worker_st *server, unsigned http_ver);
+int get_config_handler(worker_st *ws, unsigned http_ver);
+int get_string_handler(worker_st *ws, unsigned http_ver);
+int get_dl_handler(worker_st *ws, unsigned http_ver);
+
 void set_resume_db_funcs(gnutls_session_t);
 
 
